# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import re
import os
import sys
import warnings
import traceback
from contextlib import AbstractContextManager, nullcontext
from collections.abc import Sequence, Callable
from threading import RLock
from typing import Any


def unsetenv(*varnames) -> None:
    for x in varnames:
        os.environ.pop(x, None)


def hasenv(name: str) -> bool:
    return os.environ.get(name) is not None


def envint(name: str, d: int = 0) -> int:
    try:
        return int(os.environ.get(name, d))
    except ValueError:
        return d


def envbool(name: str, d: bool = False) -> bool:
    try:
        v = os.environ.get(name, "").lower()
        if v is None:
            return d
        if v in ("yes", "true", "on"):
            return True
        if v in ("no", "false", "off"):
            return False
        return bool(int(v))
    except ValueError:
        return d


def envfloat(name: str, d: float = 0) -> float:
    try:
        return float(os.environ.get(name, d))
    except ValueError:
        return d


def restore_script_env(env):
    # On OSX PythonExecWrapper sets various env vars to point into the bundle
    # and records the original variable contents. Here we revert them back
    # to their original state in case any of those changes cause problems
    # when running a command.
    if "_PYTHON_WRAPPER_VARS" in env:
        for v in env["_PYTHON_WRAPPER_VARS"].split():
            origv = "_" + v
            if env.get(origv):
                env[v] = env[origv]
            elif v in env:
                del env[v]
            del env[origv]
        del env["_PYTHON_WRAPPER_VARS"]
    return env


def shellsub(s: str, subs: dict) -> str:
    """ shell style string substitution using the dictionary given """
    if subs:
        for var, value in subs.items():
            try:
                if isinstance(s, bytes):
                    vbin = str(value).encode()
                    s = s.replace(f"${var}".encode(), vbin)
                    s = s.replace(("${%s}" % var).encode(), vbin)
                else:
                    vstr = str(value)
                    s = s.replace(f"${var}", vstr)
                    s = s.replace("${%s}" % var, vstr)
            except (TypeError, ValueError):
                msg = f"failed to substitute {var!r} with value {value!r} ({type(value)}) in {s!r}"
                raise ValueError(msg) from None
    return s


def osexpand(s: str, actual_username="", uid=0, gid=0, subs=None) -> str:
    if not s:
        return s

    def expanduser(var: str):
        if actual_username and var.startswith("~/"):
            # replace "~/" with "~$actual_username/"
            return os.path.expanduser("~%s/%s" % (actual_username, var[2:]))
        return os.path.expanduser(var)

    d = dict(subs or {})
    d |= {
        "PID": os.getpid(),
        "HOME": expanduser("~/"),
    }
    if os.name == "posix":
        d |= {
            "UID": uid or os.geteuid(),
            "GID": gid or os.getegid(),
        }
        from xpra.os_util import OSX
        if not OSX:
            from xpra.platform.posix.paths import get_runtime_dir
            rd = get_runtime_dir()
            if rd and "XDG_RUNTIME_DIR" not in os.environ:
                d["XDG_RUNTIME_DIR"] = rd
    if actual_username:
        d["USERNAME"] = actual_username
        d["USER"] = actual_username
    # first, expand the substitutions themselves,
    # as they may contain references to other variables:
    ssub = {}
    for k, v in d.items():
        ssub[k] = expanduser(shellsub(str(v), d))
    return os.path.expandvars(expanduser(shellsub(expanduser(s), ssub)))


class OSEnvContext:
    __slots__ = ("env", "kwargs")

    def __init__(self, **kwargs):
        self.env = {}
        self.kwargs = kwargs

    def __enter__(self):
        self.env = os.environ.copy()
        os.environ.update(self.kwargs)

    def __exit__(self, *_args):
        os.environ.clear()
        os.environ.update(self.env)

    def __repr__(self):
        return "OSEnvContext"


# a global dictionary used for
# showing warning message just once
_once_only = set()


def first_time(key: str) -> bool:
    if key not in _once_only:
        _once_only.add(key)
        return True
    return False


class IgnoreWarningsContext(AbstractContextManager):

    def __enter__(self):
        warnings.filterwarnings("ignore", category=DeprecationWarning)

    def __exit__(self, exc_type, exc_val, exc_tb):
        warnings.filterwarnings("default")

    def __repr__(self):
        return "IgnoreWarningsContext"


def ignorewarnings(fn: Callable, *args) -> Any:
    import warnings
    try:
        warnings.filterwarnings("ignore", category=DeprecationWarning)
        return fn(*args)
    finally:
        warnings.filterwarnings("default")


class nomodule_context:
    __slots__ = ("module_name", "saved_module")

    def __init__(self, module_name: str):
        self.module_name = module_name

    def __enter__(self):
        self.saved_module = sys.modules.get(self.module_name)
        # noinspection PyTypeChecker
        sys.modules[self.module_name] = None  # type: ignore[assignment]

    def __exit__(self, *_args):
        if sys.modules.get(self.module_name) is None:
            if self.saved_module is None:
                sys.modules.pop(self.module_name, None)
            else:
                sys.modules[self.module_name] = self.saved_module

    def __repr__(self):
        return f"nomodule_context({self.module_name})"


numpy_import_lock = RLock()


class NumpyImportContext(AbstractContextManager):

    def __init__(self, info: str, blocking=False):
        self.blocking = blocking
        self.info = info
        self.backtrace: list[str] = traceback.format_stack()

    def __enter__(self):
        if not numpy_import_lock.acquire(blocking=self.blocking):
            from xpra.log import Logger
            log = Logger("util")

            def log_backtrace(backtrace):
                for bt in backtrace:
                    for line in bt.split("\n"):
                        if line.strip():
                            log.warn(line)

            log.warn("numpy lock was already acquired from:")
            log_backtrace(self.backtrace)
            log.warn("failed to acquire it again from:")
            log_backtrace(traceback.format_stack())
            raise RuntimeError(f"the numpy import lock is already held by {self.info!r}!")
        os.environ["XPRA_NUMPY_IMPORT"] = "1"

    def __exit__(self, exc_type, exc_val, exc_tb):
        wait = envint("XPRA_NUMPY_LOCK_SLEEP", 0)
        import time
        time.sleep(wait)
        os.environ.pop("XPRA_NUMPY_IMPORT", None)
        numpy_import_lock.release()

    def __repr__(self):
        return f"numpy_import_context({self.info}, {self.blocking=})"


def numpy_import_context(subsystem: str, blocking=False) -> AbstractContextManager:
    # ie: subsystem = "OpenGL: glx context"
    key = subsystem.split(" ", 1)[0].split(":", 1)[0]
    # ie key = "OpenGL"
    env_name = "XPRA_"+(key.upper())+"_NUMPY"
    if env_name not in os.environ:
        env_name = "XPRA_NUMPY"
    allow_numpy = envbool(env_name, True)
    if allow_numpy:
        import threading
        thread = threading.current_thread()
        info = subsystem + f" in thread {thread.ident}: {thread.name!r}"
        return NumpyImportContext(info=info, blocking=blocking)
    return nomodule_context("numpy")


_saved_env = os.environ.copy()


def save_env() -> None:
    global _saved_env
    _saved_env = os.environ.copy()


def get_saved_env() -> dict[str, str]:
    return _saved_env.copy()


def get_saved_env_var(var, default="") -> str:
    return _saved_env.get(var, default)


def get_exec_env(remove: Sequence[str] = ("LS_COLORS", "LESSOPEN", "HISTCONTROL", "HISTSIZE", ),
                 keep: Sequence[str] = ()) -> dict[str, str]:
    env: dict[str, str] = {}
    for k, v in os.environ.items():
        # anything matching `remove` is dropped:
        if any(re.match(pattern, k) for pattern in remove):
            continue
        # if `keep` is empty, then we ignore it, otherwise we require a match:
        if keep and not any(re.match(pattern, k) for pattern in keep):
            continue
        # let's make things more complicated than they should be:
        # on win32, the environment can end up containing unicode, and subprocess chokes on it:
        try:
            env[k] = v.encode("utf8").decode("latin1")
        except UnicodeError:
            env[k] = str(v)
    env["XPRA_SKIP_UI"] = "1"
    env["XPRA_FORCE_COLOR_LOG"] = "1"
    return env


class SilenceWarningsContext(AbstractContextManager):

    def __init__(self, *categories):
        if sys.warnoptions:
            self.context = nullcontext()
            self.categories = ()
        else:
            self.categories = categories
            self.context = warnings.catch_warnings()

    def __enter__(self):
        self.context.__enter__()
        for category in self.categories:
            warnings.filterwarnings("ignore", category=category)

    def __exit__(self, exc_type, exc_val, exc_tb):
        warnings.filterwarnings("default")
        self.context.__exit__(exc_type, exc_val, exc_tb)

    def __repr__(self):
        return f"IgnoreWarningsContext({self.categories})"



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2023 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from dbus.types import UInt32, Int32

from xpra.os_util import gi_import
from xpra.dbus.helper import native_to_dbus
from xpra.net.common import PacketType
from xpra.platform.posix.fd_portal import REMOTEDESKTOP_IFACE
from xpra.platform.posix.fd_portal_shadow import PortalShadow
from xpra.log import Logger

Gdk = gi_import("Gdk")

log = Logger("shadow")
keylog = Logger("shadow", "keyboard")
mouselog = Logger("shadow", "mouse")


class RemoteDesktop(PortalShadow):
    def __init__(self, attrs: dict[str, str]):
        super().__init__(attrs)
        self.session_type = "pipewire desktop"
        self.input_devices = 0
        self.keymap = Gdk.Keymap.get_default()
        if not self.keymap:
            log.warn("Warning: no access to the keymap, cannot simulate key events")

    def get_server_mode(self) -> str:
        return "pipewire shadow"

    def set_keymap(self, server_source, force=False) -> None:
        keylog.info("key mapping not implemented - YMMV")

    def do_process_mouse_common(self, proto, device_id: int, wid: int, pointer, props) -> bool:
        if self.readonly or not self.input_devices:
            return False
        win = self._id_to_window.get(wid)
        if not win:
            mouselog.error(f"Error: window {wid} not found")
            return False
        x, y = pointer[:2]
        node_id = win.pipewire_id
        options = native_to_dbus([], "{sv}")
        self.portal_interface.NotifyPointerMotionAbsolute(
            self.session_handle,
            options,
            node_id,
            x, y,
            dbus_interface=REMOTEDESKTOP_IFACE)
        return True

    def do_process_button_action(self, proto, device_id: int, wid: int, button: int, pressed: bool, pointer,
                                 props) -> None:
        options = native_to_dbus([], "{sv}")
        mouselog(f"button-action: button={button}, pressed={pressed}")
        evdev_button = {
            1: 0x110,  # BTN_LEFT
            2: 0x111,  # BTN_RIGHT
            3: 0x112,  # BTN_MIDDLE
        }.get(button, -1)
        if evdev_button < 0:
            mouselog.warn(f"Warning: button {button} not recognized")
            return
        self.portal_interface.NotifyPointerButton(
            self.session_handle,
            options,
            Int32(evdev_button),
            UInt32(pressed),
            dbus_interface=REMOTEDESKTOP_IFACE)

    def _process_key_action(self, proto, packet: PacketType) -> None:
        if self.readonly or not self.input_devices or not self.keymap:
            return
        keyname = str(packet[2])
        pressed = bool(packet[3])
        keyval = int(packet[5])
        keystr = str(packet[6])
        ss = self.get_server_source(proto)
        if ss is None:
            return
        self.set_ui_driver(ss)
        keylog(f"key: name={keyname}, keyval={keyval}, keystr={keystr}")
        options = native_to_dbus([], "{sv}")
        # if keystr:
        #    keysym = Gdk.KEY_0
        #    self.portal_interface.NotifyKeyboardKeysym(
        #        self.session_handle,
        #        options,
        #        Int32(keysym),
        #        UInt32(pressed))
        # GDK code must be moved elsewhere
        ukeyname = keyname[:1].upper() + keyname[2:]
        skeyval = getattr(Gdk, f"KEY_{keyname}", 0) or getattr(Gdk, f"KEY_{ukeyname}", 0) or keyval
        entries = self.keymap.get_entries_for_keyval(skeyval)
        keylog(f"get_entries_for_keyval({skeyval})={entries}")
        if not entries or not entries[0]:
            keylog(f"no matching entries in keymap for {skeyval}")
            return
        for keymapkey in entries[1]:
            keycode = keymapkey.keycode - 8
            if keycode <= 0:
                continue
            self.portal_interface.NotifyKeyboardKeycode(
                self.session_handle,
                options,
                Int32(keycode),
                UInt32(pressed),
                dbus_interface=REMOTEDESKTOP_IFACE)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from ctypes import (
    cdll, c_size_t, c_void_p, Structure, cast, c_char, c_int, pointer, POINTER,
    WinDLL,  # @UnresolvedImport
)
from ctypes.wintypes import HDC, HANDLE, BOOL, BYTE, LPCSTR, DWORD, WORD

from xpra.platform.win32.common import CreateDCA, DeleteDC, gdi32
from xpra.log import Logger

log = Logger("printing", "win32")

CHAR = BYTE
LPHANDLE = PHANDLE = POINTER(HANDLE)
LPBYTE = POINTER(BYTE)
LPDWORD = POINTER(DWORD)
LPPRINTER_DEFAULTS = c_void_p
PSECURITY_DESCRIPTOR = HANDLE

msvcrt = cdll.msvcrt
malloc = msvcrt.malloc
malloc.argtypes = [c_size_t]
malloc.restype = c_void_p
free = msvcrt.free
free.argtypes = [c_void_p]
free.restype = None
winspool = WinDLL('winspool.drv', use_last_error=True)
OpenPrinterA = winspool.OpenPrinterA
OpenPrinterA.restype = BOOL
OpenPrinterA.argtypes = [LPCSTR, LPHANDLE, LPPRINTER_DEFAULTS]
ClosePrinter = winspool.ClosePrinter
ClosePrinter.restype = BOOL
ClosePrinter.argtypes = [HANDLE]
GetPrinterA = winspool.GetPrinterA
GetPrinterA.restype = BOOL
GetPrinterA.argtypes = [HANDLE, DWORD, c_void_p, DWORD, LPDWORD]


class DOCINFO(Structure):
    _fields_ = [
        ("cbSize", c_int),
        ("lpszDocName", LPCSTR),
        ("lpszOutput", LPCSTR),
        ("lpszDatatype", LPCSTR),
        ("fwType", DWORD),
    ]


LPDOCINFO = POINTER(DOCINFO)
StartDocA = gdi32.StartDocA  # @UndefinedVariable
StartDocA.argtypes = [HDC, LPDOCINFO]
StartDocA.restype = int
EndDoc = gdi32.EndDoc  # @UndefinedVariable
EndDoc.argtypes = [HDC]
EndDoc.restype = int
StartPage = gdi32.StartPage  # @UndefinedVariable
StartPage.argtypes = [HDC]
EndPage = gdi32.EndPage  # @UndefinedVariable
EndPage.argtypes = [HDC]
TextOutA = gdi32.TextOutA  # @UndefinedVariable
TextOutA.restype = BOOL
TextOutA.argtypes = [HDC, c_int, c_int, LPCSTR, c_int]

CCHDEVICENAME = 32


class DEVMODE(Structure):
    # noinspection PyTypeChecker
    _fields_ = [
        ("dmDeviceName", c_char * CCHDEVICENAME),
        ("dmSpecVersion", WORD),
        ("dmDriverVersion", WORD),
        ("dmSize", WORD),
        ("dmDriverExtra", WORD),
        ("dmFields", DWORD),
    ]


LPDEVMODE = POINTER(DEVMODE)


class PRINTER_INFO_1(Structure):
    _fields_ = [
        ("Flags", DWORD),
        ("pDescription", LPCSTR),
        ("pName", LPCSTR),
        ("pComment", LPCSTR),
    ]


class PRINTER_INFO_2(Structure):
    _fields_ = [
        ("pServerName", LPCSTR),
        ("pPrinterName", LPCSTR),
        ("pShareName", LPCSTR),
        ("pPortName", LPCSTR),
        ("pDriverName", LPCSTR),
        ("pComment", LPCSTR),
        ("pLocation", LPCSTR),
        ("pDevMode", LPDEVMODE),
        ("pSepFile", LPCSTR),
        ("pPrintProcessor", LPCSTR),
        ("pDatatype", LPCSTR),
        ("pParameters", LPCSTR),
        ("pSecurityDescriptor", PSECURITY_DESCRIPTOR),
        ("Attributes", DWORD),
        ("Priority", DWORD),
        ("DefaultPriority", DWORD),
        ("StartTime", DWORD),
        ("UntilTime", DWORD),
        ("Status", DWORD),
        ("cJobs", DWORD),
        ("AveragePPM", DWORD),
    ]


class PRINTER_INFO_8(Structure):
    _fields_ = [
        ("pDevMode", LPDEVMODE),
    ]


class PRINTER_INFO_9(Structure):
    _fields_ = [
        ("pDevMode", LPDEVMODE),
    ]


class GDIPrinterContext:

    def __init__(self, printer_name):
        self.printer_name = printer_name
        self.handle = None
        self.buf = None
        self.info1 = None
        self.info2 = None
        self.info8 = None
        self.info9 = None
        self.hdc = None

    def __enter__(self):
        self.handle = HANDLE()
        name = LPCSTR(self.printer_name)
        if not OpenPrinterA(name, pointer(self.handle), None):
            raise ValueError(f"failed to open printer {self.printer_name!r}")
        log("OpenPrinter: handle=%#x", self.handle.value)
        size = DWORD(0)
        GetPrinterA(self.handle, 1, None, 0, pointer(size))
        if size.value == 0:
            raise RuntimeError("GetPrinterA PRINTER_INFO_1 failed for '%s'" % self.printer_name)
        log("GetPrinter: PRINTER_INFO_1 size=%#x", size.value)
        self.info1 = malloc(size.value)
        if not GetPrinterA(self.handle, 1, self.info1, size.value, pointer(size)):
            raise RuntimeError("GetPrinterA PRINTER_INFO_1 failed for '%s'" % self.printer_name)
        info = cast(self.info1, POINTER(PRINTER_INFO_1))
        log(" flags=%#x" % info[0].Flags)
        log(" name=%#s" % info[0].pName)
        log(" description=%s" % info[0].pDescription)
        log(" comment=%s" % info[0].pComment)

        size = DWORD(0)
        GetPrinterA(self.handle, 2, None, 0, pointer(size))
        if size.value == 0:
            raise RuntimeError("GetPrinterA PRINTER_INFO_2 failed for '%s'" % self.printer_name)
        log("GetPrinter: PRINTER_INFO_2 size=%#x", size.value)
        self.info2 = malloc(size.value)
        if GetPrinterA(self.handle, 2, self.info2, size.value, pointer(size)):
            info = cast(self.info2, POINTER(PRINTER_INFO_2))
            log(" driver=%#s" % info[0].pDriverName)

        devmode = 0
        size = DWORD(0)
        GetPrinterA(self.handle, 8, None, 0, pointer(size))
        if size.value == 0:
            raise RuntimeError("GetPrinter: PRINTER_INFO_8 failed for '%s'" % self.printer_name)
        self.info8 = malloc(size.value)
        if GetPrinterA(self.handle, 8, self.info8, size.value, pointer(size)):
            info = cast(self.info8, POINTER(PRINTER_INFO_8))
            if info[0] and info[0].pDevMode:
                devmode = cast(info[0].pDevMode, POINTER(DEVMODE))
                log("PRINTER_INFO_8: devmode=%s" % devmode)
                log("PRINTER_INFO_8: device name='%s'" % devmode[0].dmDeviceName)

        size = DWORD(0)
        GetPrinterA(self.handle, 9, None, 0, pointer(size))
        if size.value == 0:
            raise RuntimeError("GetPrinter: PRINTER_INFO_9 failed for '%s'" % self.printer_name)
        log("GetPrinter: PRINTER_INFO_9 size=%#x" % size.value)
        self.info9 = malloc(size.value)
        if GetPrinterA(self.handle, 9, self.info9, size.value, pointer(size)):
            info = cast(self.info9, POINTER(PRINTER_INFO_9))
            if info[0] and info[0].pDevMode:
                devmode = cast(info[0].pDevMode, POINTER(DEVMODE))
                log("PRINTER_INFO_9: devmode=%s" % devmode)
                log("PRINTER_INFO_9: device name=%s" % devmode[0].dmDeviceName)
        if not devmode:
            raise RuntimeError(f"failed to query a DEVMODE for {self.printer_name!r}")
        self.hdc = CreateDCA(None, name, None, devmode)
        log("CreateDCA(..)=%#x", self.hdc)
        return self.hdc

    def __exit__(self, exc_type, exc_val, exc_tb):
        log("GDIPrintContext(%s).exit%s hdc=%s, info=%s, handle=%s", self.printer_name, (exc_type, exc_val, exc_tb),
            self.hdc, (self.info1, self.info2, self.info8, self.info9), self.handle)
        if self.hdc:
            DeleteDC(self.hdc)
            self.hdc = None
        if self.info1:
            free(self.info1)
            self.info1 = None
        if self.info2:
            free(self.info2)
            self.info2 = None
        if self.info8:
            free(self.info8)
            self.info8 = None
        if self.info9:
            free(self.info9)
            self.info9 = None
        if self.handle:
            ClosePrinter(self.handle)
            self.handle = None

    def __repr__(self):
        return "GDIPrinterContext(%s)" % self.printer_name


def main(argv):
    import datetime
    title = ""
    if len(argv) == 1:
        from xpra.platform.win32.printing import get_printers
        printers = get_printers()
        log("printers: %s", printers)
        printer_name = printers.keys()[0]
    elif len(argv) == 2:
        printer_name = argv[1]
    elif len(argv) == 3:
        printer_name = argv[1]
        title = argv[2]
    else:
        log.error("usage: %s [printer-name [title]]", argv[0])
        return 1

    with GDIPrinterContext(printer_name) as hdc:
        log("hdc=%s", hdc)
        docinfo = DOCINFO()
        docinfo.lpszDocName = LPCSTR("%s\0" % title)
        log("StartDocA(%#x, %s)", hdc, docinfo)
        r = StartDocA(hdc, pointer(docinfo))
        if r < 0:
            log.error("StartDocA failed: %i", r)
            return r
        log("StartDocA()=%i" % r)
        r = StartPage(hdc)
        if r < 0:
            log.error("StartPage failed: %i", r)
            return r
        x, y = 100, 100
        s = "Test Page printed at %s" % (datetime.datetime.now())
        if not TextOutA(hdc, x, y, LPCSTR(s.encode()), len(s)):
            log.error("TextOutA failed")
            return 1
        r = EndPage(hdc)
        if r < 0:
            log.error("EndPage failed: %i", r)
            return r
        r = EndDoc(hdc)
        if r < 0:
            log.error("EndDoc failed: %i" % r)
            return r
        log("EndDoc()=%i" % r)
        return 0


if __name__ == "__main__":
    import sys

    sys.exit(main(sys.argv))



# This file is part of Xpra.
# Copyright (C) 2010 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from collections.abc import Callable, Sequence

from xpra.os_util import gi_import
from xpra.platform.keyboard_base import KeyboardBase, log
from xpra.platform.darwin.menu import getOSXMenuHelper

Gdk = gi_import("Gdk")

NUM_LOCK_KEYCODE = 71  # HARDCODED!
# a key and the keys we want to translate it into when swapping keys
# (in a list with the best options first)
KEYS_TRANSLATION_OPTIONS: dict[str, list[str]] = {
    # try to swap with "Meta" first, fallback to "Alt":
    "Control_L": ["Meta_L", "Meta_R", "Alt_L", "Alt_R"],
    "Control_R": ["Meta_R", "Meta_L", "Alt_R", "Alt_L"],
    # "Meta" always swapped with "Control"
    "Meta_L": ["Control_L", "Control_R"],
    "Meta_R": ["Control_R", "Control_L"],
    # "Alt" to "Super" (or "Hyper") so we can distinguish it from "Meta":
    "Alt_L": ["Super_L", "Super_R", "Hyper_L", "Hyper_R"],
    "Alt_R": ["Super_R", "Super_L", "Hyper_R", "Hyper_L"],
}
# keys we always want to swap,
# irrespective of the swap-keys option:
ALWAYS_SWAP: list[str] = os.environ.get("XPRA_MACOS_KEYS_ALWAYS_SWAP", "Alt_L,Alt_R").split(",")

# data extracted from:
# https://support.apple.com/en-us/HT201794
# "How to identify keyboard localizations"
# maps Apple's names into standard X11 keyboard identifiers

APPLE_LAYOUTS: dict[str, str] = {
    "Arabic": "ar",
    "Belgian": "be",
    "Bulgarian": "bg",
    "Croatian": "cr",
    "Czech": "cz",
    "Danish": "dk",
    "Dutch": "nl",
    "British": "gb",
    "US": "us",
    "Finnish": "fi",
    "Swedish": "se",
    "French": "fr",
    "German": "de",
    "Greek": "gr",
    "Hungarian": "hu",
    # "Icelandic" : "is",
    "Israel": "il",
    "Italian": "it",
    "Japanese": "jp",
    "Korean": "ko",
    "Norwegian": "no",
    "Portugese": "po",
    "Romanian": "ro",
    "Russian": "ru",
    "Slovak": "sl",
    "Spanish": "es",
    # "Swiss"     : "ch",
    "Taiwanese": "tw",
    "Thai": "th",
    "Turkey": "tr",
}


class Keyboard(KeyboardBase):
    """
        Switch Meta and Control
    """

    def init_vars(self) -> None:
        super().init_vars()
        self.swap_keys = True
        self.meta_modifier = ""
        self.control_modifier = "control"
        self.super_modifier = ""
        self.hyper_modifier = ""
        self.num_lock_modifier = ""
        self.num_lock_state = True
        self.num_lock_keycode = NUM_LOCK_KEYCODE
        self.key_translations: dict[str, tuple[int, str]] = {}

    def __repr__(self):
        return "darwin.Keyboard"

    def get_all_x11_layouts(self) -> dict[str, str]:
        x11_layouts: dict[str, str] = {}
        for name, layout in APPLE_LAYOUTS.items():
            x11_layouts[layout] = name
        return x11_layouts

    def get_layout_spec(self) -> tuple[str, str, Sequence[str], str, Sequence[str], str]:
        model = ""
        layout = "us"
        layouts = ["us"]
        variant = ""
        variants = []
        options = ""
        try:
            from AppKit import NSTextInputContext, NSTextView
            text_input_context = NSTextInputContext.alloc()
            view = NSTextView.new()
            text_input_context.initWithClient_(view)
            current_keyboard = text_input_context.selectedKeyboardInputSource()
            code = APPLE_LAYOUTS.get(current_keyboard.split(".")[-1])
            log("get_layout_spec() current_keyboard=%s, code=%s", current_keyboard, code)
            all_keyboards = text_input_context.keyboardInputSources()
            log("get_layout_spec() other keyboards=%s", repr(all_keyboards))
            if code:
                layout = code
            if all_keyboards:
                layouts = []
                for k in all_keyboards:
                    code = APPLE_LAYOUTS.get(k.split(".")[-1])
                    if code:
                        layouts.append(code)
                if not layouts:
                    layouts.append("us")
            else:
                if code and code not in layouts:
                    layouts.insert(0, code)
            log("get_layout_spec() view=%s, input_context=%s, layout=%s, layouts=%s",
                repr(view), text_input_context, layout, layouts)
        except Exception as e:
            log("get_layout_spec()", exc_info=True)
            log.error("Error querying keyboard layout:")
            log.estr(e)
        return model, layout, layouts, variant, variants, options

    def get_keymap_modifiers(self) -> tuple[dict, list[str], list[str]]:
        """
            Override superclass so we can tell the server
            that 'control' will also be missing from non key events modifiers
        """
        return {}, [], ["lock", "control"]

    def set_modifier_mappings(self, mappings: dict[str, str]) -> None:
        super().set_modifier_mappings(mappings)
        self.meta_modifier = self.modifier_keys.get("Meta_L") or self.modifier_keys.get("Meta_R") or ""
        self.control_modifier = self.modifier_keys.get("Control_L") or self.modifier_keys.get("Control_R") or "control"
        self.super_modifier = self.modifier_keys.get("Super_L") or self.modifier_keys.get("Super_R") or ""
        self.hyper_modifier = self.modifier_keys.get("Hyper_L") or self.modifier_keys.get("Hyper_R") or ""
        self.num_lock_modifier = self.modifier_keys.get("Num_Lock") or ""
        log("set_modifier_mappings(%s) meta=%s, control=%s, super=%s, hyper=%s, numlock=%s", mappings,
            self.meta_modifier, self.control_modifier, self.super_modifier, self.hyper_modifier, self.num_lock_modifier)
        # find the keysyms and keycodes to use for each key we may translate:
        for orig_keysym, keysyms in KEYS_TRANSLATION_OPTIONS.items():
            new_def = self.find_translation(keysyms)
            if new_def is not None:
                self.key_translations[orig_keysym] = new_def
        log("set_modifier_mappings(..) swap keys translations=%s", self.key_translations)

    def find_translation(self, keysyms):
        log("find_translation(%s)", keysyms)
        new_def = None
        # ie: keysyms : ["Meta_L", "Alt_L"]
        for keysym in keysyms:
            # ie: "Alt_L":
            keycodes_defs = self.modifier_keycodes.get(keysym)
            log("modifier_keycodes(%s)=%s", keysym, keycodes_defs)
            if not keycodes_defs:
                # keysym not found
                continue
            # ie: [(55, 'Alt_L'), (58, 'Alt_L'), 'Alt_L']
            for keycode_def in keycodes_defs:
                if isinstance(keycode_def, str):  # ie: 'Alt_L'
                    # no keycode found, but better than nothing:
                    new_def = 0, keycode_def  # ie: (0, 'Alt_L')
                    continue
                # an int alone is the keycode:
                if isinstance(keycode_def, int):
                    if keycode_def > 0:
                        # exact match, use it:
                        return keycode_def, keysym
                    new_def = 0, keysym
                    continue
                # below is for compatibility with older servers,
                # (we may be able to remove some of this code already)
                # look for a tuple of (keycode, keysym):
                if not isinstance(keycode_def, (list, tuple)):
                    continue
                if len(keycode_def) != 2:
                    continue
                if not (isinstance(keycode_def[0], int) and isinstance(keycode_def[1], str)):
                    continue
                if keycode_def[0] == 0:
                    new_def = keycode_def
                    continue
                # found a valid keycode, use this one:
                return keycode_def  # ie: (55, 'Alt_L')
        return new_def

    def mask_to_names(self, mask) -> list[str]:
        if self.swap_keys:
            control = self.meta_modifier
            meta = self.control_modifier
        else:
            meta = self.meta_modifier
            control = self.control_modifier
        modmap = {
            Gdk.ModifierType.SHIFT_MASK: "shift",
            Gdk.ModifierType.LOCK_MASK: "lock",
            Gdk.ModifierType.SUPER_MASK: self.super_modifier,
            Gdk.ModifierType.HYPER_MASK: self.hyper_modifier,
            Gdk.ModifierType.META_MASK: meta,
            Gdk.ModifierType.CONTROL_MASK: control,
        }
        names = []
        for modmask, modname in modmap.items():
            if (mask & modmask) and modname:
                names.append(modname)
        # don't trust GTK with numlock:
        if self.num_lock_modifier:
            if self.num_lock_state and self.num_lock_modifier not in names:
                names.append(self.num_lock_modifier)
            elif not self.num_lock_state and self.num_lock_modifier in names:
                names.remove(self.num_lock_modifier)
        log("mask_to_names(%s)=%s swap_keys=%s, modmap=%s, num_lock_state=%s, num_lock_modifier=%s",
            mask, names, self.swap_keys, modmap, self.num_lock_state, self.num_lock_modifier)
        return names

    def process_key_event(self, send_key_action_cb: Callable, wid: int, key_event) -> None:
        if self.swap_keys or key_event.keyname in ALWAYS_SWAP:
            trans = self.key_translations.get(key_event.keyname)
            if trans:
                log("swap keys: translating key '%s' to %s", key_event, trans)
                key_event.keycode, key_event.keyname = trans
        if key_event.keycode == self.num_lock_keycode:
            if not key_event.pressed:
                log("toggling numlock")
                self.num_lock_state = not self.num_lock_state
                getOSXMenuHelper().update_numlock(self.num_lock_state)
            # do not forward the "Escape" key that numlock usually comes up as
            return
        send_key_action_cb(wid, key_event)



# This file is part of Xpra.
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

#cython: boundscheck=False, wraparound=False, initializedcheck=False, always_allow_keywords=False

from libc.stdint cimport uint32_t, uintptr_t   # pylint: disable=syntax-error
from xpra.buffers.membuf cimport getbuf, MemBuf, buffer_context

from xpra.common import SizedBuffer


def hybi_unmask(data, unsigned int offset, unsigned int datalen) -> memoryview:
    cdef uintptr_t mp
    cdef unsigned int min_len = offset + 4 + datalen
    with buffer_context(data) as bc:
        if len(bc)<(<Py_ssize_t> min_len):
            raise ValueError(f"buffer too small {len(bc)} vs {min_len}: offset={offset}, datalen={datalen}")
        mp = (<uintptr_t> int(bc)) + offset
        return do_hybi_mask(mp, mp + 4, datalen)


def hybi_mask(mask: bytes, data: SizedBuffer) -> memoryview:
    if len(mask) != 4:
        raise ValueError(f"invalid mask buffer size: {len(mask)} bytes")
    with buffer_context(mask) as mbc:
        with buffer_context(data) as dbc:
            return do_hybi_mask(<uintptr_t> int(mbc), <uintptr_t> int(dbc), len(dbc))


cdef memoryview do_hybi_mask(uintptr_t mp, uintptr_t dp, unsigned int datalen):
    # we skip the first 'align' bytes in the output buffer,
    # to ensure that its alignment is the same as the input data buffer
    cdef unsigned int align = (<uintptr_t> dp) & 0x3
    cdef unsigned int initial_chars = 4 - align
    cdef MemBuf out_buf = getbuf(datalen + align, 0)
    cdef uintptr_t op = <uintptr_t> out_buf.get_mem()
    # char pointers:
    cdef unsigned char *mcbuf = <unsigned char *> mp
    cdef unsigned char *dcbuf = <unsigned char *> dp
    cdef unsigned char *ocbuf = <unsigned char *> op
    cdef unsigned int j
    # bytes at a time until we reach the 32-bit boundary:
    for i in range(initial_chars):
        ocbuf[align + i] = dcbuf[i] ^ mcbuf[i & 0x3]
    # 32-bit pointers:
    cdef uint32_t *dbuf
    cdef uint32_t *obuf
    cdef uint32_t mask_value
    cdef unsigned int uint32_steps
    cdef unsigned int last_chars
    if datalen > initial_chars:
        uint32_steps = (datalen - initial_chars) // 4
        if uint32_steps:
            dbuf = <uint32_t*> (dp + initial_chars)
            obuf = <uint32_t*> (op + align + initial_chars)
            mask_value = 0
            for i in range(4):
                mask_value = mask_value << 8
                mask_value += mcbuf[(3 - i + initial_chars) & 0x3]
            for i in range(uint32_steps):
                obuf[i] = dbuf[i] ^ mask_value
        # bytes at a time again at the end:
        last_chars = (datalen - initial_chars) & 0x3
        for i in range(last_chars):
            j = datalen - last_chars + i
            ocbuf[align + j] = dcbuf[j] ^ mcbuf[j & 0x3]
    if align > 0:
        return (memoryview(out_buf)[align:]).toreadonly()
    return memoryview(out_buf).toreadonly()



# This file is part of Xpra.
# Copyright (C) 2023 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any
from queue import Queue, Empty, Full
from collections.abc import Sequence

from xpra.os_util import gi_import
from xpra.util.env import envbool
from xpra.util.objects import typedict, AtomicInteger
from xpra.gstreamer.common import (
    import_gst, GST_FLOW_OK, get_element_str,
    get_default_appsink_attributes,
    get_caps_str,
)
from xpra.gtk.gobject import n_arg_signal
from xpra.gstreamer.pipeline import Pipeline
from xpra.codecs.constants import get_profile, CSC_ALIAS, VideoSpec
from xpra.codecs.gstreamer.common import (
    get_version, get_type, get_info,
    get_video_encoder_caps, get_video_encoder_options,
    get_gst_encoding, get_encoder_info, get_gst_rgb_format,
)
from xpra.codecs.video import getVideoHelper
from xpra.codecs.image import ImageWrapper
from xpra.log import Logger

Gst = import_gst()
log = Logger("encoder", "gstreamer")

GObject = gi_import("GObject")

log(f"capture: {get_type()} {get_version()}")

SAVE_TO_FILE = envbool("XPRA_SAVE_TO_FILE")

generation = AtomicInteger()


class Capture(Pipeline):
    """
    Uses a GStreamer pipeline to capture the screen
    """
    __gsignals__ = Pipeline.__generic_signals__.copy()
    __gsignals__["new-image"] = n_arg_signal(3)

    def __init__(self, element: str = "ximagesrc", pixel_format: str = "BGRX",
                 width: int = 0, height: int = 0, framerate: int = 0):
        super().__init__()
        self.capture_element = element.split(" ")[0]
        self.pixel_format: str = pixel_format
        self.width: int = width
        self.height: int = height
        self.frames: int = 0
        self.framerate: int = framerate
        self.image: Queue[ImageWrapper] = Queue(maxsize=1)
        self.sink = None
        self.capture = None
        self.extra_client_info = {}
        self.create_pipeline(element)
        assert width > 0 and height > 0

    def create_pipeline(self, capture_element: str = "ximagesrc") -> None:
        elements = [
            f"{capture_element} name=capture",  # ie: ximagesrc or pipewiresrc
        ]
        if self.framerate > 0:
            elements.append(f"video/x-raw,framerate={self.framerate}/1")
        elements += [
            "videoconvert",
            "appsink name=sink emit-signals=true max-buffers=10 drop=true sync=false async=false qos=false",
        ]
        if not self.setup_pipeline_and_bus(elements):
            raise RuntimeError("failed to setup gstreamer pipeline")
        self.sink = self.pipeline.get_by_name("sink")
        self.capture = self.pipeline.get_by_name("capture")

        def sh(sig, handler):
            self.element_connect(self.sink, sig, handler)

        sh("new-sample", self.on_new_sample)
        sh("new-preroll", self.on_new_preroll)

    def on_new_sample(self, _bus) -> int:
        sample = self.sink.emit("pull-sample")
        buf = sample.get_buffer()
        size = buf.get_size()
        log("on_new_sample size=%s", size)
        if size:
            data = buf.extract_dup(0, size)
            self.frames += 1
            bytesperpixel = len(self.pixel_format)
            rowstride = self.width * bytesperpixel
            image = ImageWrapper(x=0, y=0, width=self.width, height=self.height, pixels=data,
                                 pixel_format=self.pixel_format, depth=24,
                                 rowstride=rowstride, bytesperpixel=bytesperpixel, planes=ImageWrapper.PACKED)
            log(f"image={image}")
            while True:
                try:
                    self.image.get(block=False)
                except Empty:
                    break
            try:
                self.image.put_nowait(image)
            except Full:
                log("image queue is already full")
            else:
                client_info = {"frame": self.frames}
                self.emit("new-image", self.pixel_format, image, client_info)
        return GST_FLOW_OK

    def on_new_preroll(self, _appsink) -> int:
        log("new-preroll")
        return GST_FLOW_OK

    def get_image(self, x: int = 0, y: int = 0, width: int = 0, height: int = 0) -> ImageWrapper | None:
        log("get_image%s", (x, y, width, height))
        if self.state == "stopped":
            return None
        try:
            return self.image.get(timeout=5)
        except Empty:
            return None

    def refresh(self) -> bool:
        return not self.image.empty()

    def clean(self) -> None:
        self.stop()
        f = self.file
        if f:
            self.file = None
            f.close()

    def get_type(self) -> str:
        return self.capture_element


GObject.type_register(Capture)


def choose_video_encoder(preferred_encoding: str, full_csc_modes: typedict) -> VideoSpec | None:
    log(f"choose_video_encoder({preferred_encoding}, {full_csc_modes})")
    vh = getVideoHelper()
    scores: dict[int, list[VideoSpec]] = {}
    for encoding in full_csc_modes.keys():
        csc_modes = full_csc_modes.strtupleget(encoding)
        encoder_specs = vh.get_encoder_specs(encoding)
        # ignore the input colorspace for now,
        # and assume we can `videoconvert` to this format
        # TODO: take csc into account for scoring
        for codec_list in encoder_specs.values():
            for codec in codec_list:
                if not codec.codec_type.startswith("gstreamer"):
                    continue
                if "*" not in csc_modes:
                    # verify that the client can decode this csc mode:
                    matches = tuple(x for x in codec.output_colorspaces if x in csc_modes)
                    if not matches:
                        log(f"skipped {codec}: {codec.output_colorspaces} not in {csc_modes}")
                        continue
                # prefer encoding matching what the user requested and GPU accelerated:
                gpu = bool(codec.gpu_cost > codec.cpu_cost)
                score = int(codec.encoding == preferred_encoding or gpu) * 100 + codec.quality + codec.score_boost
                log(f"score({codec.codec_type})={score} ({gpu=}, quality={codec.quality}, boost={codec.score_boost})")
                # (lowest score wins)
                scores.setdefault(-score, []).append(codec)
    log(f"choose_video_encoder: scores={scores}")
    if not scores:
        return None
    best_score = sorted(scores)[0]
    best = scores[best_score][0]
    log(f"choose_video_encoder({preferred_encoding}, {full_csc_modes})={best}")
    return best


def choose_csc(modes: Sequence[str], quality=100) -> str:
    prefer = "YUV420P" if quality < 80 else "YUV444P"
    if not modes or prefer in modes:
        return prefer
    return modes[0]


class CaptureAndEncode(Capture):
    """
    Uses a GStreamer pipeline to capture the screen
    and encode it to a video stream
    """

    def __init__(self, element: str = "ximagesrc", encoding="vp8", encoder="vp8enc",
                 pixel_format: str = "YUV420P", width: int = 0, height: int = 0, framerate: int = 0):
        self.encoding = encoding
        self.encoder = encoder
        super().__init__(element, pixel_format, width, height, framerate)

    def create_pipeline(self, capture_element: str = "ximagesrc") -> None:
        options = typedict({
            "speed": 100,
            "quality": 100,
        })
        einfo = get_encoder_info(self.encoder)
        log(f"{self.encoder}: {einfo=}")
        self.profile = get_profile(options, self.encoding, csc_mode=self.pixel_format,
                                   default_profile="high" if self.encoder == "x264enc" else "")
        eopts = get_video_encoder_options(self.encoder, self.profile, options)
        vcaps = get_video_encoder_caps(self.encoder)
        self.extra_client_info: dict[str, Any] = vcaps.copy()
        if self.profile:
            vcaps["profile"] = self.profile
            self.extra_client_info["profile"] = self.profile
        gst_encoding = get_gst_encoding(self.encoding)  # ie: "hevc" -> "video/x-h265"
        elements = [
            f"{capture_element} name=capture",  # ie: ximagesrc or pipewiresrc
        ]
        if self.framerate > 0:
            elements.append(f"video/x-raw,framerate={self.framerate}/1")
        elements += [
            "queue leaky=2 max-size-buffers=1",
            "videoconvert",
            "video/x-raw,format=%s" % get_gst_rgb_format(self.pixel_format),
            get_element_str(self.encoder, eopts),
            get_caps_str(gst_encoding, vcaps),
            get_element_str("appsink", get_default_appsink_attributes()),
        ]
        if not self.setup_pipeline_and_bus(elements):
            raise RuntimeError("failed to setup gstreamer pipeline")
        self.sink: Gst.Element = self.pipeline.get_by_name("sink")
        self.capture = self.pipeline.get_by_name("capture")
        self.file = None

        def sh(sig, handler):
            self.element_connect(self.sink, sig, handler)

        sh("new-sample", self.on_new_sample)
        sh("new-preroll", self.on_new_preroll)

    def on_new_sample(self, _bus) -> int:
        sample = self.sink.emit("pull-sample")
        buf = sample.get_buffer()
        size = buf.get_size()
        log("on_new_sample size=%s", size)
        if size:
            data = buf.extract_dup(0, size)
            self.frames += 1
            client_info = self.extra_client_info
            client_info["frame"] = self.frames
            client_info["csc"] = CSC_ALIAS.get(self.pixel_format, self.pixel_format)
            self.extra_client_info = {}
            self.emit("new-image", self.encoding, data, client_info)
            if SAVE_TO_FILE:
                if not self.file:
                    gen = generation.increase()
                    filename = "gstreamer-" + str(gen) + f".{self.encoding}"
                    self.file = open(filename, "wb")
                    log.info(f"saving gstreamer {self.encoding} stream to {filename!r}")
                self.file.write(data)
        return GST_FLOW_OK

    def on_new_preroll(self, _appsink) -> int:
        log("new-preroll")
        return GST_FLOW_OK

    def refresh(self) -> bool:
        return True

    def clean(self) -> None:
        self.stop()

    def get_type(self) -> str:
        return f"{self.capture_element}-{self.pixel_format}"


GObject.type_register(CaptureAndEncode)


def capture_and_encode(capture_element: str, encoding: str,
                       full_csc_modes: typedict, w: int, h: int, framerate=0) -> CaptureAndEncode | None:
    encoder_spec = choose_video_encoder(encoding, full_csc_modes)
    if not encoder_spec:
        log(f"unable to find a GStreamer video encoder with csc modes={full_csc_modes}")
        return None
    assert encoder_spec.codec_type.startswith("gstreamer-")
    encoder = encoder_spec.codec_type[len("gstreamer-"):]
    encoding = encoder_spec.encoding
    csc_mode = encoder_spec.input_colorspace
    return CaptureAndEncode(capture_element, encoding, encoder, csc_mode, w, h, framerate)


def selftest(_full=False) -> None:
    log("gstreamer encoder selftest: %s", get_info())
    GLib = gi_import("GLib")
    from xpra.gtk.util import get_root_size
    w, h = get_root_size()
    c = Capture(width=w, height=h)
    loop = GLib.MainLoop()

    def check() -> bool:
        i = c.get_image()
        if i:
            c.stop()
            return False
        return True

    GLib.timeout_add(500, check)
    GLib.timeout_add(2000, c.stop)
    GLib.timeout_add(2500, loop.quit)
    c.start()
    loop.run()


if __name__ == "__main__":
    selftest()



# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import socket
from typing import Callable, Any
from collections.abc import Sequence

from xpra.os_util import OSX, WIN32
from xpra.util.env import envbool


def import_zeroconf() -> tuple[Any, Callable]:
    from xpra.net.mdns.zeroconf_publisher import ZeroconfPublishers, get_interface_index
    return ZeroconfPublishers, get_interface_index


def import_avahi() -> tuple[Any, Callable]:
    from xpra.net.mdns.avahi_publisher import AvahiPublishers, get_interface_index
    return AvahiPublishers, get_interface_index


MDNS_WARNING = False


def mdns_publish(display_name: str, listen_on, text_dict=None) -> Sequence:
    global MDNS_WARNING
    if MDNS_WARNING is True:
        return ()
    from xpra.log import Logger
    log = Logger("mdns")
    log("mdns_publish%s", (display_name, listen_on, text_dict))
    try:
        from xpra.net import mdns
        assert mdns
        from xpra.net.mdns import XPRA_TCP_MDNS_TYPE, XPRA_UDP_MDNS_TYPE, RFB_MDNS_TYPE
    except ImportError as e:
        log(f"mdns support is not installed: {e}")
        return ()
    PREFER_ZEROCONF = envbool("XPRA_PREFER_ZEROCONF", True)
    imports = (import_zeroconf, import_avahi)
    if not PREFER_ZEROCONF:
        imports = (import_avahi, import_zeroconf)
    MDNSPublishers = get_interface_index = None
    exceptions = []
    for i in imports:
        try:
            MDNSPublishers, get_interface_index = i()
            break
        except ImportError as e:
            log("mdns import failure", exc_info=True)
            exceptions.append(e)
    if not MDNSPublishers or not get_interface_index:
        MDNS_WARNING = True
        log.warn("Warning: failed to load the mdns publishers")
        for exc in exceptions:
            log.warn(" %s", str(exc) or type(exc))
        log.warn(" install 'python-avahi', 'python-zeroconf'")
        log.warn(" or use the 'mdns=no' option")
        return ()
    d = dict(text_dict or {})
    # ensure we don't have duplicate interfaces:
    f_listen_on = {}
    for host, port in listen_on:
        f_listen_on[(get_interface_index(host), port)] = (host, port)
    try:
        name = socket.gethostname()
    except OSError:
        name = "Xpra"
    if display_name and not (OSX or WIN32):
        name += f" {display_name}"
    mode = d.get("mode", "tcp")
    service_type = {
        "rfb": RFB_MDNS_TYPE,
        "quic": XPRA_UDP_MDNS_TYPE,
        "webtransport": XPRA_UDP_MDNS_TYPE,
    }.get(mode, XPRA_TCP_MDNS_TYPE)
    index = 0
    aps = []
    for host, port in listen_on:
        sn = name
        mode_str = mode
        if index > 0:
            mode_str = f"{mode}-{index + 1}"
        if mode not in ("tcp", "rfb"):
            sn += f" ({mode_str})"
        listen = ((host, port),)
        index += 1
        aps.append(MDNSPublishers(listen, sn, service_type=service_type, text_dict=d))
    return aps



# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
import os.path
from typing import Any
from collections.abc import Sequence

from xpra.exit_codes import ExitCode
from xpra.os_util import WIN32, POSIX, OSX, is_admin
from xpra.util.io import load_binary_file, umask_context
from xpra.scripts.config import InitExit, InitException, TRUE_OPTIONS
from xpra.util.env import osexpand, envbool
from xpra.util.parsing import parse_encoded_bin_data
from xpra.util.str_fn import print_nested_dict, csv, Ellipsizer, std

SSL_RETRY = envbool("XPRA_SSL_RETRY", True)

SSL_ATTRIBUTES: Sequence[str] = (
    "cert", "key", "ca-certs", "ca-data",
    "protocol",
    "client-verify-mode", "server-verify-mode", "verify-flags",
    "check-hostname", "server-hostname",
    "options", "ciphers",
)

KEY_SIZE = 4096
KEY_DAYS = 3650
KEY_SUBJ = "/C=US/ST=Denial/L=Springfield/O=Dis/CN=localhost"

KEY_FILENAME = "key.pem"
CERT_FILENAME = "cert.pem"
SSL_CERT_FILENAME = "ssl-cert.pem"

logger = None


def get_ssl_logger():
    global logger
    if not logger:
        from xpra.log import Logger
        logger = Logger("network", "ssl")
    return logger


def get_ssl_attributes(opts, server_side: bool = True, overrides: dict | None = None) -> dict[str, Any]:
    args: dict[str, Any] = {
        "server-side": server_side,
    }
    for attr in SSL_ATTRIBUTES:
        v = (overrides or {}).get(attr)
        if v is None and opts:
            fn = attr.replace("-", "_")
            ssl_attr = f"ssl_{fn}"  # ie: "ssl_ca_certs"
            v = getattr(opts, ssl_attr)
        args[attr] = v
    return args


def find_ssl_cert(filename: str = SSL_CERT_FILENAME) -> str:
    ssllog = get_ssl_logger()
    # try to locate the cert file from known locations
    from xpra.platform.paths import get_ssl_cert_dirs  # pylint: disable=import-outside-toplevel
    dirs = get_ssl_cert_dirs()
    ssllog(f"find_ssl_cert({filename}) get_ssl_cert_dirs()={dirs}")
    for d in dirs:
        p = osexpand(d)
        if not os.path.exists(p):
            ssllog(f"ssl cert dir {p!r} does not exist")
            continue
        f = os.path.join(p, filename)
        if not os.path.exists(f):
            ssllog(f"ssl cert {f!r} does not exist")
            continue
        if not os.path.isfile(f):
            ssllog.warn(f"Warning: {f!r} is not a file")
            continue
        if not os.access(f, os.R_OK):
            ssllog.info(f"SSL certificate file {f!r} is not accessible")
            continue
        ssllog(f"found ssl cert {f!r}")
        return os.path.abspath(f)
    return ""


def ssl_wrap_socket(sock, **kwargs):
    context, wrap_kwargs = get_ssl_wrap_socket_context(**kwargs)
    ssllog = get_ssl_logger()
    ssllog("ssl_wrap_socket(%s, %s) context=%s, wrap_kwargs=%s", sock, kwargs, context, wrap_kwargs)
    return do_wrap_socket(sock, context, **wrap_kwargs)


def log_ssl_info(ssl_sock) -> None:
    ssllog = get_ssl_logger()
    ssllog("server_hostname=%s", ssl_sock.server_hostname)
    cipher = ssl_sock.cipher()
    if cipher:
        ssllog.info(" %s, %s bits", cipher[0], cipher[2])
    try:
        cert = ssl_sock.getpeercert()
    except ValueError:
        pass
    else:
        if cert:
            ssllog("certificate:")
            print_nested_dict(ssl_sock.getpeercert(), prefix=" ", print_fn=ssllog)


SSL_VERIFY_EXPIRED = 10
SSL_VERIFY_WRONG_HOST = 20
SSL_VERIFY_SELF_SIGNED = 18
SSL_VERIFY_UNTRUSTED_ROOT = 19
SSL_VERIFY_IP_MISMATCH = 64
SSL_VERIFY_HOSTNAME_MISMATCH = 62
SSL_VERIFY_CODES: dict[int, str] = {
    SSL_VERIFY_EXPIRED: "expired",  # also revoked!
    SSL_VERIFY_WRONG_HOST: "wrong host",
    SSL_VERIFY_SELF_SIGNED: "self-signed",
    SSL_VERIFY_UNTRUSTED_ROOT: "untrusted-root",
    SSL_VERIFY_IP_MISMATCH: "ip-mismatch",
    SSL_VERIFY_HOSTNAME_MISMATCH: "hostname-mismatch",
}


class SSLVerifyFailure(InitExit):
    def __init__(self, status, msg, verify_code, ssl_sock):
        super().__init__(status, msg)
        self.verify_code = verify_code
        self.ssl_sock = ssl_sock


def ssl_handshake(ssl_sock) -> None:
    ssllog = get_ssl_logger()
    try:
        ssl_sock.do_handshake(True)
        ssllog.info("SSL handshake complete, %s", ssl_sock.version())
        log_ssl_info(ssl_sock)
    except Exception as e:
        ssllog("do_handshake", exc_info=True)
        log_ssl_info(ssl_sock)
        import ssl
        ssleof_error = getattr(ssl, "SSLEOFError", None)
        if ssleof_error and isinstance(e, ssleof_error):
            return
        status = ExitCode.SSL_FAILURE
        ssl_cert_verification_error = getattr(ssl, "SSLCertVerificationError", None)
        if ssl_cert_verification_error and isinstance(e, ssl_cert_verification_error):
            verify_code = getattr(e, "verify_code", 0)
            ssllog("verify_code=%s", SSL_VERIFY_CODES.get(verify_code, verify_code))
            try:
                msg = getattr(e, "verify_message") or (e.args[1].split(":", 2)[2])
            except (ValueError, IndexError):
                msg = str(e)
            status = ExitCode.SSL_CERTIFICATE_VERIFY_FAILURE
            ssllog("host failed SSL verification: %s", msg)
            raise SSLVerifyFailure(status, msg, verify_code, ssl_sock) from None
        raise InitExit(status, f"SSL handshake failed: {e}") from None


def get_ssl_verify_mode(verify_mode_str: str):
    # parse verify-mode:
    import ssl
    ssl_cert_reqs = getattr(ssl, "CERT_" + verify_mode_str.upper(), None)
    if ssl_cert_reqs is None:
        values = [k[len("CERT_"):].lower() for k in dir(ssl) if k.startswith("CERT_")]
        raise InitException(f"invalid ssl-server-verify-mode {verify_mode_str!r}, must be one of: " + csv(values))
    return ssl_cert_reqs


def get_ssl_wrap_socket_context(cert="", key="", key_password="", ca_certs="", ca_data="",
                                protocol: str = "TLS",
                                client_verify_mode: str = "optional", server_verify_mode: str = "required",
                                verify_flags: str = "X509_STRICT",
                                check_hostname: bool = False, server_hostname="",
                                options: str = "ALL,NO_COMPRESSION", ciphers: str = "DEFAULT",
                                server_side: bool = True):
    if server_side and not cert:
        cert = find_ssl_cert(SSL_CERT_FILENAME)
        if not cert:
            raise InitException("you must specify an 'ssl-cert' file to use ssl sockets")
    ssllog = get_ssl_logger()
    ssllog("get_ssl_wrap_socket_context%s", (
        cert, key, ca_certs, ca_data,
        protocol,
        client_verify_mode, server_verify_mode,
        verify_flags,
        check_hostname, server_hostname,
        options, ciphers,
        server_side)
    )
    if server_side:
        ssl_cert_reqs = get_ssl_verify_mode(client_verify_mode)
    else:
        ssl_cert_reqs = get_ssl_verify_mode(server_verify_mode)
    ssllog(" verify_mode for server_side=%s : %s", server_side, ssl_cert_reqs)
    # ca-certs:
    if ca_certs == "default":
        ca_certs = ""
    elif ca_certs == "auto":
        ca_certs = find_ssl_cert("ca-cert.pem")
    ssllog(" ca-certs=%s", ca_certs)
    # parse protocol:
    import ssl
    if protocol.upper() == "TLS":
        protocol = "TLS_SERVER" if server_side else "TLS_CLIENT"
    proto = getattr(ssl, "PROTOCOL_" + protocol.upper().replace("TLSV", "TLSv"), None)
    if proto is None:
        values = [k[len("PROTOCOL_"):] for k in dir(ssl) if k.startswith("PROTOCOL_")]
        raise InitException(f"invalid ssl-protocol {protocol!r}, must be one of: " + csv(values))
    ssllog(" protocol=%#x", proto)
    # ca_data may be hex encoded:
    ca_data = parse_encoded_bin_data(ca_data or "")
    ssllog(" cadata=%s", Ellipsizer(ca_data))

    kwargs: dict[str, bool | str] = {
        "server_side": server_side,
        "do_handshake_on_connect": False,
        "suppress_ragged_eofs": True,
    }
    # parse ssl-verify-flags as CSV:
    ssl_verify_flags = 0
    for x in verify_flags.split(","):
        x = x.strip()
        if not x:
            continue
        v = getattr(ssl, "VERIFY_" + x.upper(), None)
        if v is None:
            raise InitException(f"invalid ssl verify-flag: {x!r}")
        ssl_verify_flags |= v
    ssllog(" verify-flags=%#x", ssl_verify_flags)
    # parse ssl-options as CSV:
    ssl_options = 0
    for x in options.split(","):
        x = x.strip()
        if not x:
            continue
        v = getattr(ssl, "OP_" + x.upper(), None)
        if v is None:
            raise InitException(f"invalid ssl option: {x!r}")
        ssl_options |= v
    ssllog(" options=%#x", ssl_options)

    context = ssl.SSLContext(proto)
    context.set_ciphers(ciphers)
    if not server_side:
        context.check_hostname = check_hostname
    context.verify_mode = ssl_cert_reqs
    context.verify_flags = ssl_verify_flags
    context.options = ssl_options
    ssllog(" cert=%s, key=%s", cert, key)
    if cert:
        if cert == "auto":
            # try to locate the cert file from known locations
            cert = find_ssl_cert()
            if not cert:
                raise InitException("failed to automatically locate an SSL certificate to use")
        key_password = key_password or os.environ.get("XPRA_SSL_KEY_PASSWORD")
        ssllog("context.load_cert_chain%s", (cert or None, key or None, key_password))
        try:
            # we must pass a `None` value to ignore `keyfile`:
            context.load_cert_chain(certfile=cert, keyfile=key or None, password=key_password)
        except ssl.SSLError as e:
            ssllog("load_cert_chain", exc_info=True)
            raise InitException(f"SSL error, failed to load certificate chain: {e}") from e
    # if not server_side and (check_hostname or (ca_certs and ca_certs.lower()!="default")):
    if not server_side:
        kwargs["server_hostname"] = server_hostname
    if ssl_cert_reqs != ssl.CERT_NONE:
        if server_side:
            purpose = ssl.Purpose.CLIENT_AUTH
        else:
            purpose = ssl.Purpose.SERVER_AUTH
            ssllog(" check_hostname=%s, server_hostname=%s", check_hostname, server_hostname)
            if context.check_hostname and not server_hostname:
                raise InitException("ssl error: check-hostname is set but server-hostname is not")
        ssllog(" load_default_certs(%s)", purpose)
        context.load_default_certs(purpose)

        if not ca_certs or ca_certs.lower() == "default":
            ssllog(" using default certs")
            # load_default_certs already calls set_default_verify_paths()
        elif not os.path.exists(ca_certs):
            raise InitException(f"invalid ssl-ca-certs file or directory: {ca_certs}")
        elif os.path.isdir(ca_certs):
            ssllog(" loading ca certs from directory '%s'", ca_certs)
            context.load_verify_locations(capath=ca_certs)
        else:
            ssllog(" loading ca certs from file '%s'", ca_certs)
            if not os.path.isfile(ca_certs):
                raise InitException(f"{ca_certs!r} is not a valid ca file")
            context.load_verify_locations(cafile=ca_certs)
        if ca_data:
            context.load_verify_locations(cadata=ca_data)
    elif check_hostname and not server_side:
        ssllog("cannot check hostname client side with verify mode %s", ssl_cert_reqs)
    return context, kwargs


def do_wrap_socket(tcp_socket, context, **kwargs):
    wrap_socket = context.wrap_socket
    assert tcp_socket
    ssllog = get_ssl_logger()
    ssllog("do_wrap_socket(%s, %s, %s)", tcp_socket, context, kwargs)
    import ssl
    if WIN32:
        # on win32, setting the tcp socket to blocking doesn't work?
        # we still hit the following errors that we need to retry:
        from xpra.net import bytestreams
        bytestreams.CAN_RETRY_EXCEPTIONS = (ssl.SSLWantReadError, ssl.SSLWantWriteError)
    else:
        tcp_socket.setblocking(True)
    try:
        ssl_sock = wrap_socket(tcp_socket, **kwargs)
    except (InitExit, InitException):
        ssllog.debug("wrap_socket(%s, %s)", tcp_socket, kwargs, exc_info=True)
        raise
    except Exception as e:
        ssllog.debug("wrap_socket(%s, %s)", tcp_socket, kwargs, exc_info=True)
        ssleof_error = getattr(ssl, "SSLEOFError", None)
        if ssleof_error and isinstance(e, ssleof_error):
            return None
        raise InitExit(ExitCode.SSL_FAILURE, f"Cannot wrap socket {tcp_socket}: {e}") from None
    return ssl_sock


def ssl_retry(e, ssl_ca_certs: str) -> dict[str, Any]:
    ssllog = get_ssl_logger()
    ssllog("ssl_retry(%s, %s) SSL_RETRY=%s", e, ssl_ca_certs, SSL_RETRY)
    if not SSL_RETRY:
        return {}
    if not isinstance(e, SSLVerifyFailure):
        return {}
    # we may be able to ask the user if he wants to accept this certificate
    verify_code = e.verify_code
    ssl_sock = e.ssl_sock
    msg = str(e)
    del e
    addr = ssl_sock.getpeername()[:2]
    port = addr[-1]
    server_hostname = ssl_sock.server_hostname
    ssllog("ssl_retry: peername=%s, server_hostname=%s", addr, server_hostname)
    if verify_code not in (
            SSL_VERIFY_SELF_SIGNED, SSL_VERIFY_WRONG_HOST,
            SSL_VERIFY_IP_MISMATCH, SSL_VERIFY_HOSTNAME_MISMATCH,
    ):
        ssllog("ssl_retry: %s not handled here", SSL_VERIFY_CODES.get(verify_code, verify_code))
        return {}
    if not server_hostname:
        ssllog("ssl_retry: no server hostname")
        return {}
    ssllog("ssl_retry: server_hostname=%s, ssl verify_code=%s (%i)",
           server_hostname, SSL_VERIFY_CODES.get(verify_code, verify_code), verify_code)

    def confirm(*args) -> bool:
        from xpra.scripts import pinentry
        ret = pinentry.confirm(*args)
        ssllog("run_pinentry_confirm(..) returned %r", ret)
        return ret

    options = load_ssl_options(server_hostname, port)
    # self-signed cert:
    if verify_code == SSL_VERIFY_SELF_SIGNED:
        if ssl_ca_certs not in ("", "default"):
            ssllog("self-signed cert does not match %r", ssl_ca_certs)
            return {}
        # perhaps we already have the certificate for this hostname
        cert_file = find_ssl_config_file(server_hostname, port, CERT_FILENAME)
        if cert_file:
            ssllog("retrying with %r", cert_file)
            options["ca-certs"] = cert_file
            return options
        # download the certificate data
        import ssl
        try:
            cert_data = ssl.get_server_certificate(addr)
        except ssl.SSLError:
            cert_data = None
        if not cert_data:
            ssllog.warn("Warning: failed to get server certificate from %s", addr)
            return {}
        ssllog("downloaded ssl cert data for %s: %s", addr, Ellipsizer(cert_data))
        # ask the user if he wants to accept this certificate:
        title = "SSL Certificate Verification Failure"
        prompt = "Do you want to accept this certificate?"
        if not confirm((msg,), title, prompt):
            return {}
        filename = save_ssl_config_file(server_hostname, port,
                                        CERT_FILENAME, "certificate", cert_data.encode("latin1"))
        if not filename:
            ssllog.warn("Warning: failed to save certificate data")
            return {}
        options["ca-certs"] = filename
        save_ssl_options(server_hostname, port, options)
        return options
    if verify_code in (SSL_VERIFY_WRONG_HOST, SSL_VERIFY_IP_MISMATCH, SSL_VERIFY_HOSTNAME_MISMATCH):
        # ask the user if he wants to skip verifying the host
        title = "SSL Certificate Verification Failure"
        prompt = "Do you want to connect anyway?"
        r = confirm((msg,), title, prompt)
        ssllog("run_pinentry_confirm(..) returned %r", r)
        if r:
            ssllog.info(title)
            ssllog.info(" user chose to connect anyway")
            ssllog.info(" retrying without checking the hostname")
            options["check-hostname"] = False
            save_ssl_options(server_hostname, port, options)
            return options
    return {}


def load_ssl_options(server_hostname: str, port: int) -> dict[str, Any]:
    ssllog = get_ssl_logger()
    f = find_ssl_config_file(server_hostname, port, "options")
    options = {}
    if f:
        try:
            with open(f, encoding="utf8") as fd:
                for line in fd.readlines():
                    line = line.rstrip("\n\r")
                    if not line or line.startswith("#"):
                        continue
                    parts = line.split("=", 1)
                    if len(parts) != 2:
                        continue
                    k, v = parts
                    if k not in SSL_ATTRIBUTES:
                        ssllog("Warning: unknown SSL attribute %r in %r", k, f)
                        continue
                    # some options use boolean values, convert them back:
                    options[k] = (v.lower() in TRUE_OPTIONS) if k in ("check-hostname",) else v
        except OSError as e:
            ssllog.warn("Warning: failed to read %r: %s", f, e)
    ssllog("load_ssl_options%s=%s (from %r)", (server_hostname, port), options, f)
    return options


def save_ssl_options(server_hostname: str, port, options: dict) -> str:
    ssllog = get_ssl_logger()
    boptions = b"\n".join(("{}={}".format(k.replace("_", "-"), v)).encode("latin1") for k, v in options.items())
    boptions += b"\n"
    f = save_ssl_config_file(server_hostname, port,
                             "options", "configuration options", boptions)
    ssllog("save_ssl_options%s saved to %r", (server_hostname, port, options), f)
    return f


def find_ssl_config_file(hostname: str, port=443, filename=CERT_FILENAME) -> str:
    return do_find_ssl_config_file(hostname, port, filename) or do_find_ssl_config_file(hostname, 0, filename)


def do_find_ssl_config_file(server_hostname: str, port=443, filename=CERT_FILENAME) -> str:
    ssllog = get_ssl_logger()
    from xpra.platform.paths import get_ssl_hosts_config_dirs
    dirs = get_ssl_hosts_config_dirs()
    host_dirname = std(server_hostname, extras="-.:#_")
    if port:
        host_dirname += f"_{port}"
    host_dirs = [os.path.join(osexpand(d), host_dirname) for d in dirs]
    ssllog(f"looking for {filename!r} in {host_dirs}")
    for d in host_dirs:
        f = os.path.join(d, filename)
        if os.path.exists(f):
            ssllog(f"found {f}")
            return os.path.abspath(f)
    return ""


def save_ssl_config_file(server_hostname: str, port=443,
                         filename=CERT_FILENAME, fileinfo="certificate", filedata=b"") -> str:
    ssllog = get_ssl_logger()
    from xpra.platform.paths import get_ssl_hosts_config_dirs
    dirs = get_ssl_hosts_config_dirs()
    host_dirname = std(server_hostname, extras="-.:#_")
    if port:
        host_dirname += f"_{port}"
    host_dirs = [os.path.join(osexpand(d), host_dirname) for d in dirs]
    ssllog(f"save_ssl_config_file%s dirs={dirs}, host_dirname={host_dirname}, host_dirs={host_dirs}",
           (server_hostname, port, filename, fileinfo, Ellipsizer(filedata)), )
    # if there is an existing host config dir, try to use it:
    for d in [x for x in host_dirs if os.path.exists(x)]:
        f = os.path.join(d, filename)
        try:
            with open(f, "wb") as fd:
                fd.write(filedata)
            ssllog.info(f"saved SSL {fileinfo} to {f!r}")
            return f
        except OSError:
            ssllog(f"failed to save SSL {fileinfo} to {f!r}", exc_info=True)
    # try to create a host config dir:
    for d in host_dirs:
        folders = os.path.normpath(d).split(os.sep)
        # we have to be careful and create the 'ssl' dir with 0o700 permissions
        # but any directory above that can use 0o755
        try:
            ssl_dir_index = len(folders) - 1
            while ssl_dir_index > 0 and folders[ssl_dir_index] != "ssl":
                ssl_dir_index -= 1
            if ssl_dir_index > 1:
                parent = os.path.join(*folders[:ssl_dir_index - 1])
                ssl_dir = os.path.join(*folders[:ssl_dir_index])
                os.makedirs(parent, exist_ok=True)
                os.makedirs(ssl_dir, mode=0o700, exist_ok=True)
            os.makedirs(d, mode=0o700)
            f = os.path.join(d, filename)
            with open(f, "wb") as fd:
                fd.write(filedata)
            ssllog.info(f"saved SSL {fileinfo} to {f!r}")
            return f
        except OSError:
            ssllog(f"failed to save cert data to {d!r}", exc_info=True)
    return ""


def gen_ssl_cert() -> tuple[str, str]:
    log = get_ssl_logger()
    keypath = find_ssl_cert(KEY_FILENAME)
    certpath = find_ssl_cert(CERT_FILENAME)
    if keypath and certpath:
        log.info("found an existing SSL certificate:")
        log.info(f" {keypath!r}")
        log.info(f" {certpath!r}")
        return keypath, certpath
    from shutil import which
    openssl = which("openssl") or os.environ.get("OPENSSL", "")
    if not openssl:
        raise InitExit(ExitCode.SSL_FAILURE, "cannot find openssl executable")
    openssl_config = ""
    creationflags = 0
    if WIN32:
        from xpra.platform.paths import get_app_dir
        from subprocess import CREATE_NO_WINDOW
        creationflags = CREATE_NO_WINDOW
        openssl_config = os.path.join(get_app_dir(), "etc", "ssl", "openssl.cnf")
    if is_admin():
        # running as root, use global location:
        if OSX:
            xpra_dir = "/Library/Application Support/Xpra"
        elif WIN32:
            from xpra.platform.win32.paths import get_program_data_dir
            xpra_dir = os.path.join(get_program_data_dir(), "Xpra")
        else:
            prefix = "" if sys.prefix == "/usr" else sys.prefix
            xpra_dir = f"{prefix}/etc/xpra"
        ssldir = f"{xpra_dir}/ssl"
        if not os.path.exists(xpra_dir):
            os.mkdir(ssldir, 0o777)
        if not os.path.exists(ssldir):
            os.mkdir(ssldir, 0o777)
    else:
        from xpra.platform.paths import get_ssl_cert_dirs
        dirs = [d for d in get_ssl_cert_dirs() if not d.startswith("/etc") and not d.startswith("/usr") and d != "./"]
        # use the first writeable one:
        log(f"testing ssl dirs: {dirs}")
        ssldir = ""
        for sdir in dirs:
            path = osexpand(sdir)
            if os.path.exists(path) and os.path.isdir(path) and os.access(path, os.W_OK):
                ssldir = path
                log(f"found writeable ssl dir {ssldir!r}")
                break
        if not ssldir:
            # we may have to create the parent directories:
            log("no ssl dir found, trying to create one")
            for sdir in dirs:
                path = osexpand(sdir)
                head = path
                create_dirs = []
                while head:
                    head, tail = os.path.split(head)        #ie: "/home/user/.config/xpra", "ssl"
                    log(f"{head=}, {tail=}")
                    if tail:
                        create_dirs.append(tail)
                    else:
                        break
                    if tail.find("xpra") >= 0:
                        # don't create directories above our one
                        break
                if create_dirs and head:
                    while create_dirs:
                        tail = create_dirs.pop()
                        head = os.path.join(head, tail)
                        try:
                            os.mkdir(head, 0o777)
                        except OSError as e:
                            log(f"failed to create {head!r} for {path!r}: {e}")
                            continue
                    ssldir = path
                    log(f"created {ssldir!r}")
                    break
    keypath = f"{ssldir}/{KEY_FILENAME}"
    certpath = f"{ssldir}/{CERT_FILENAME}"
    cmd = [
        openssl,
        "req", "-new",
        "-newkey", f"rsa:{KEY_SIZE}",
        "-days", f"{KEY_DAYS}",
        "-nodes", "-x509",
        "-subj", KEY_SUBJ,
        "-keyout", keypath,
        "-out", certpath,
    ]
    if openssl_config and os.path.exists(openssl_config):
        cmd += ["-config", openssl_config]
    log.info("generating a new SSL certificate:")
    log.info(f" {keypath!r}")
    log.info(f" {certpath!r}")
    log(f"openssl command: {cmd}")
    from subprocess import Popen
    with umask_context(0o022):
        with Popen(cmd, creationflags=creationflags) as p:
            exit_code = p.wait()
    if exit_code != 0:
        raise InitExit(ExitCode.FAILURE, f"openssl command returned {exit_code}")
    key = load_binary_file(keypath)
    cert = load_binary_file(certpath)
    sslcert = key+cert
    sslcertpath = f"{ssldir}/{SSL_CERT_FILENAME}"
    with open(sslcertpath, "wb") as f:
        if POSIX:
            os.fchmod(f.fileno(), 0o600)
        f.write(sslcert)
    return keypath, certpath


def strip_cert(data: bytes) -> bytes:
    BEGIN = b"-----BEGIN CERTIFICATE-----"
    if data.find(BEGIN) >= 0:
        data = BEGIN + data.split(BEGIN, 1)[1]
    END = b"-----END CERTIFICATE-----"
    if data.find(END) > 0:
        data = data.split(END, 1)[0] + END + b"\n"
    return data



# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os

from xpra.challenge.handler import AuthenticationHandler
from xpra.util.io import load_binary_file
from xpra.log import Logger

log = Logger("auth")


class Handler(AuthenticationHandler):

    def __init__(self, **kwargs):
        if "filename" in kwargs:
            self.password_file: str = kwargs.get("filename", "")
            return
        if "password-files" in kwargs:
            files = kwargs.get("password-files", [])
            if files:
                self.password_file = files.pop(0)
                return
        self.password_file = ""

    def __repr__(self):
        return "file"

    def get_digest(self) -> str:
        return ""

    def handle(self, challenge: bytes, digest: str, prompt: str) -> bytes:  # pylint: disable=unused-argument
        log("handle(..) password_file=%s", self.password_file)
        if not self.password_file:
            return b""
        filename = os.path.expanduser(self.password_file)
        data = load_binary_file(filename)
        log("loaded password data from %s: %s", filename, bool(data))
        return data



# This file is part of Xpra.
# Copyright (C) 2025 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from libc.stddef cimport wchar_t
from libc.stdint cimport uint8_t, uint16_t, uint32_t, int32_t, int64_t, uint64_t

ctypedef void *amf_handle
ctypedef long amf_long
ctypedef uint8_t amf_bool
ctypedef uint8_t amf_uint8
ctypedef uint16_t amf_uint16
ctypedef uint32_t amf_uint32
ctypedef int32_t amf_int32
ctypedef int64_t amf_int64
ctypedef int64_t amf_pts
ctypedef size_t amf_size
ctypedef void* AMFDataAllocatorCB
ctypedef void* AMFComponentOptimizationCallback


cdef extern from "Python.h":
    object PyUnicode_FromWideChar(wchar_t *w, Py_ssize_t size)


cdef extern from "string.h":
    size_t wcslen(const wchar_t *str)


cdef extern from "stdarg.h":
    ctypedef struct va_list:
        pass


cdef extern from "core/Result.h":
    enum AMF_RESULT:
        AMF_OK
        AMF_FAIL
        # common errors
        AMF_UNEXPECTED
        AMF_ACCESS_DENIED
        AMF_INVALID_ARG
        AMF_OUT_OF_RANGE
        AMF_OUT_OF_MEMORY
        AMF_INVALID_POINTER
        AMF_NO_INTERFACE
        AMF_NOT_IMPLEMENTED
        AMF_NOT_SUPPORTED
        AMF_NOT_FOUND
        AMF_ALREADY_INITIALIZED
        AMF_NOT_INITIALIZED
        AMF_INVALID_FORMAT
        AMF_WRONG_STATE
        AMF_FILE_NOT_OPEN
        # device common codes
        AMF_NO_DEVICE
        AMF_DIRECTX_FAILED
        AMF_OPENCL_FAILED
        AMF_GLX_FAILED
        AMF_XV_FAILED
        AMF_ALSA_FAILED
        # result codes
        AMF_EOF
        AMF_REPEAT
        AMF_INPUT_FULL
        AMF_RESOLUTION_CHANGED
        AMF_RESOLUTION_UPDATED
        # error codes
        AMF_INVALID_DATA_TYPE
        AMF_INVALID_RESOLUTION
        AMF_CODEC_NOT_SUPPORTED
        AMF_SURFACE_FORMAT_NOT_SUPPORTED
        AMF_SURFACE_MUST_BE_SHARED
        # component video decoder
        AMF_DECODER_NOT_PRESENT
        AMF_DECODER_SURFACE_ALLOCATION_FAILED
        AMF_DECODER_NO_FREE_SURFACES
        # component video encoder
        AMF_ENCODER_NOT_PRESENT
        # component dem
        AMF_DEM_ERROR
        AMF_DEM_PROPERTY_READONLY
        AMF_DEM_REMOTE_DISPLAY_CREATE_FAILED
        AMF_DEM_START_ENCODING_FAILED
        AMF_DEM_QUERY_OUTPUT_FAILED
        # component TAN
        AMF_TAN_CLIPPING_WAS_REQUIRED           # Resulting data was truncated to meet output type's value limits.
        AMF_TAN_UNSUPPORTED_VERSION             # Not supported version requested, solely for TANCreateContext().
        AMF_NEED_MORE_INPUT                     # returned by AMFComponent::SubmitInput did not produce a buffer because more input submissions are required.
        # device vulkan
        AMF_VULKAN_FAILED


cdef inline MEMORY_TYPE_STR(AMF_MEMORY_TYPE memory):
    return {
        AMF_MEMORY_UNKNOWN: "UNKNOWN",
        AMF_MEMORY_HOST: "HOST",
        AMF_MEMORY_DX9: "DX9",
        AMF_MEMORY_DX11: "DX11",
        AMF_MEMORY_OPENCL: "OPENCL",
        AMF_MEMORY_OPENGL: "OPENGL",
        AMF_MEMORY_XV: "XV",
        AMF_MEMORY_GRALLOC: "GRALLOC",
        AMF_MEMORY_COMPUTE_FOR_DX9: "COMPUTE_FOR_DX9",
        AMF_MEMORY_COMPUTE_FOR_DX11: "COMPUTE_FOR_DX11",
        AMF_MEMORY_VULKAN: "VULKAN",
        AMF_MEMORY_DX12: "DX12",
    }.get(memory, "unknown")


cdef inline RESULT_STR(AMF_RESULT res):
    return {
        AMF_OK: "OK",
        AMF_FAIL: "FAIL",
        AMF_UNEXPECTED: "UNEXPECTED",
        AMF_ACCESS_DENIED: "ACCESS_DENIED",
        AMF_INVALID_ARG: "INVALID_ARG",
        AMF_OUT_OF_RANGE: "OUT_OF_RANGE",
        AMF_OUT_OF_MEMORY: "OUT_OF_MEMORY",
        AMF_INVALID_POINTER: "INVALID_POINTER",
        AMF_NO_INTERFACE: "NO_INTERFACE",
        AMF_NOT_IMPLEMENTED: "NOT_IMPLEMENTED",
        AMF_NOT_SUPPORTED: "NOT_SUPPORTED",
        AMF_NOT_FOUND: "NOT_FOUND",
        AMF_ALREADY_INITIALIZED: "ALREADY_INITIALIZED",
        AMF_NOT_INITIALIZED: "NOT_INITIALIZED",
        AMF_INVALID_FORMAT: "INVALID_FORMAT",
        AMF_WRONG_STATE: "WRONG_STATE",
        AMF_FILE_NOT_OPEN: "FILE_NOT_OPEN",
        AMF_NO_DEVICE: "NO_DEVICE",
        AMF_DIRECTX_FAILED: "DIRECTX_FAILED",
        AMF_OPENCL_FAILED: "OPENCL_FAILED",
        AMF_GLX_FAILED: "GLX_FAILED",
        AMF_XV_FAILED: "XV_FAILED",
        AMF_ALSA_FAILED: "ALSA_FAILED",
        AMF_EOF: "EOF",
        AMF_REPEAT: "REPEAT",
        AMF_INPUT_FULL: "INPUT_FULL",
        AMF_RESOLUTION_CHANGED: "RESOLUTION_CHANGED",
        AMF_RESOLUTION_UPDATED: "RESOLUTION_UPDATED",
        AMF_INVALID_DATA_TYPE: "INVALID_DATA_TYPE",
        AMF_INVALID_RESOLUTION: "INVALID_RESOLUTION",
        AMF_CODEC_NOT_SUPPORTED: "CODEC_NOT_SUPPORTED",
        AMF_SURFACE_FORMAT_NOT_SUPPORTED: "SURFACE_FORMAT_NOT_SUPPORTED",
        AMF_SURFACE_MUST_BE_SHARED: "SURFACE_MUST_BE_SHARED",
        AMF_DECODER_NOT_PRESENT: "DECODER_NOT_PRESENT",
        AMF_DECODER_SURFACE_ALLOCATION_FAILED: "DECODER_SURFACE_ALLOCATION_FAILED",
        AMF_DECODER_NO_FREE_SURFACES: "DECODER_NO_FREE_SURFACES",
        AMF_ENCODER_NOT_PRESENT: "ENCODER_NOT_PRESENT",
        AMF_DEM_ERROR: "DEM_ERROR",
        AMF_DEM_PROPERTY_READONLY: "DEM_PROPERTY_READONLY",
        AMF_DEM_REMOTE_DISPLAY_CREATE_FAILED: "DEM_REMOTE_DISPLAY_CREATE_FAILED",
        AMF_DEM_START_ENCODING_FAILED: "DEM_START_ENCODING_FAILED",
        AMF_DEM_QUERY_OUTPUT_FAILED: "DEM_QUERY_OUTPUT_FAILED",
        AMF_TAN_CLIPPING_WAS_REQUIRED: "TAN_CLIPPING_WAS_REQUIRED",
        AMF_TAN_UNSUPPORTED_VERSION: "TAN_UNSUPPORTED_VERSION",
        AMF_NEED_MORE_INPUT: "NEED_MORE_INPUT",
        AMF_VULKAN_FAILED: "VULKAN_FAILED",
    }.get(res, "")


cdef extern from "core/Data.h":
    ctypedef enum AMF_DATA_TYPE:
        AMF_DATA_BUFFER         # 0
        AMF_DATA_SURFACE        # 1
        AMF_DATA_AUDIO_BUFFER   # 2
        AMF_DATA_USER           # 1000

    ctypedef enum AMF_MEMORY_TYPE:
        AMF_MEMORY_UNKNOWN          # 0
        AMF_MEMORY_HOST             # 1
        AMF_MEMORY_DX9              # 2
        AMF_MEMORY_DX11             # 3
        AMF_MEMORY_OPENCL           # 4
        AMF_MEMORY_OPENGL           # 5
        AMF_MEMORY_XV               # 6
        AMF_MEMORY_GRALLOC          # 7
        AMF_MEMORY_COMPUTE_FOR_DX9  # 8  deprecated, the same as AMF_MEMORY_OPENCL
        AMF_MEMORY_COMPUTE_FOR_DX11 # 9 deprecated, the same as AMF_MEMORY_OPENCL
        AMF_MEMORY_VULKAN           # 10
        AMF_MEMORY_DX12             # 11

    ctypedef amf_long (*DATA_ACQUIRE)(AMFData* pThis)
    ctypedef amf_long (*DATA_RELEASE)(AMFData* pThis)
    ctypedef AMF_RESULT (*DATA_QUERYINTERFACE)(AMFData* pThis, const AMFGuid *interfaceID, void** ppInterface)
    ctypedef amf_size (*DATA_GETPROPERTYCOUNT)(AMFData* pThis)
    ctypedef AMF_MEMORY_TYPE (*DATA_GETMEMORYTYPE)(AMFData* pThis)
    ctypedef AMF_DATA_TYPE (*DATA_GETDATATYPE)(AMFData* pThis)
    ctypedef AMF_RESULT (*DATA_CONVERT)(AMFData* pThis, AMF_MEMORY_TYPE type)
    ctypedef AMF_RESULT (*DATA_INTEROP)(AMFData* pThis, AMF_MEMORY_TYPE type)

    ctypedef struct AMFDataVtbl:
        DATA_ACQUIRE Acquire
        DATA_RELEASE Release
        DATA_QUERYINTERFACE QueryInterface
        DATA_GETPROPERTYCOUNT GetPropertyCount
        DATA_GETMEMORYTYPE GetMemoryType
        DATA_GETDATATYPE GetDataType
        DATA_CONVERT Convert
        DATA_INTEROP Interop

    ctypedef struct AMFData:
        const AMFDataVtbl *pVtbl


cdef extern from "core/Platform.h":
    ctypedef struct AMFSize:
        amf_int32 width
        amf_int32 height

    ctypedef struct AMFRate:
        amf_uint32 num
        amf_uint32 den

    ctypedef struct AMFGuid:
        amf_uint32 data1
        amf_uint16 data2
        amf_uint16 data3
        amf_uint8 data41
        amf_uint8 data42
        amf_uint8 data43
        amf_uint8 data44
        amf_uint8 data45
        amf_uint8 data46
        amf_uint8 data47
        amf_uint8 data48
    ctypedef struct AMFRect:
        amf_int32 left
        amf_int32 top
        amf_int32 right
        amf_int32 bottom


cdef extern from "core/Variant.h":
    ctypedef enum AMF_VARIANT_TYPE:
        AMF_VARIANT_EMPTY           # 0
        AMF_VARIANT_BOOL            # 1
        AMF_VARIANT_INT64           # 2
        AMF_VARIANT_DOUBLE          # 3
        AMF_VARIANT_RECT            # 4
        AMF_VARIANT_SIZE            # 5
        AMF_VARIANT_POINT           # 6
        AMF_VARIANT_RATE            # 7
        AMF_VARIANT_RATIO           # 8
        AMF_VARIANT_COLOR           # 9
        AMF_VARIANT_STRING          # 10  // value is char*
        AMF_VARIANT_WSTRING         # 11  // value is wchar_t*
        AMF_VARIANT_INTERFACE       # 12  // value is AMFInterface*
        AMF_VARIANT_FLOAT           # 13
        AMF_VARIANT_FLOAT_SIZE      # 14
        AMF_VARIANT_FLOAT_POINT2D   # 15
        AMF_VARIANT_FLOAT_POINT3D   # 16
        AMF_VARIANT_FLOAT_VECTOR4D  # 17


    ctypedef struct AMFVariantStruct:
        AMF_VARIANT_TYPE type
        AMFSize sizeValue
        AMFRate rateValue
        int64_t int64Value
        amf_bool boolValue

    AMF_RESULT AMFVariantInit(AMFVariantStruct* pDest)
    AMF_RESULT AMFVariantAssignInt64(AMFVariantStruct* pDest, amf_int64 value)
    AMF_RESULT AMFVariantAssignSize(AMFVariantStruct* pDest, const AMFSize &value)
    AMF_RESULT AMFVariantAssignRate(AMFVariantStruct* pDest, const AMFRate &value)
    AMF_RESULT AMFVariantClear(AMFVariantStruct* pDest)


cdef extern from "core/Plane.h":
    ctypedef enum AMF_PLANE_TYPE:
        AMF_PLANE_UNKNOWN       # 0
        AMF_PLANE_PACKED        # 1 for all packed formats: BGRA, YUY2, etc
        AMF_PLANE_Y             # 2
        AMF_PLANE_UV            # 3
        AMF_PLANE_U             # 4
        AMF_PLANE_V             # 5

    ctypedef amf_long (*PLANE_ACQUIRE)(AMFPlane* pThis)
    ctypedef amf_long (*PLANE_RELEASE)(AMFPlane* pThis)
    ctypedef AMF_RESULT (*PLANE_QUERYINTERFACE)(AMFPlane* pThis, const AMFGuid *interfaceID, void** ppInterface)

    ctypedef AMF_PLANE_TYPE (*PLANE_GETTYPE)(AMFPlane* pThis)
    ctypedef void* (*PLANE_GETNATIVE)(AMFPlane* pThis)
    ctypedef amf_int32 (*PLANE_GETPIXELSIZEINBYTES)(AMFPlane* pThis)
    ctypedef amf_int32 (*PLANE_GETOFFSETX)(AMFPlane* pThis)
    ctypedef amf_int32 (*PLANE_GETOFFSETY)(AMFPlane* pThis)
    ctypedef amf_int32 (*PLANE_GETWIDTH)(AMFPlane* pThis)
    ctypedef amf_int32 (*PLANE_GETHEIGHT)(AMFPlane* pThis)
    ctypedef amf_int32 (*PLANE_GETHPITCH)(AMFPlane* pThis)
    ctypedef amf_int32 (*PLANE_GETVPITCH)(AMFPlane* pThis)
    ctypedef amf_bool (*PLANE_ISTILED)(AMFPlane* pThis)

    ctypedef struct AMFPlaneVtbl:
        PLANE_ACQUIRE Acquire
        PLANE_RELEASE Release
        PLANE_QUERYINTERFACE QueryInterface

        PLANE_GETTYPE GetType
        PLANE_GETNATIVE GetNative
        PLANE_GETPIXELSIZEINBYTES GetPixelSizeInBytes
        PLANE_GETOFFSETX GetOffsetX
        PLANE_GETOFFSETY GetOffsetY
        PLANE_GETWIDTH GetWidth
        PLANE_GETHEIGHT GetHeight
        PLANE_GETHPITCH GetHPitch
        PLANE_GETVPITCH GetVPitch
        PLANE_ISTILED IsTiled

    ctypedef struct AMFPlane:
        const AMFPlaneVtbl *pVtbl


cdef extern from "core/Buffer.h":
    enum AMF_BUFFER_USAGE_BITS:
        AMF_BUFFER_USAGE_DEFAULT
        AMF_BUFFER_USAGE_NONE
        AMF_BUFFER_USAGE_CONSTANT
        AMF_BUFFER_USAGE_SHADER_RESOURCE
        AMF_BUFFER_USAGE_UNORDERED_ACCESS
        AMF_BUFFER_USAGE_TRANSFER_SRC
        AMF_BUFFER_USAGE_TRANSFER_DST
        AMF_BUFFER_USAGE_NOSYNC
        AMF_BUFFER_USAGE_DECODER_SRC

    ctypedef amf_long (*BUFFER_ACQUIRE)(AMFBuffer* pThis)
    ctypedef amf_long (*BUFFER_RELEASE)(AMFBuffer* pThis)
    ctypedef AMF_RESULT (*BUFFER_QUERYINTERFACE)(AMFBuffer* pThis, const AMFGuid *interfaceID, void** ppInterface)
    ctypedef AMF_RESULT (*BUFFER_SETPROPERTY)(AMFBuffer* pThis, const wchar_t *pName, AMFVariantStruct value)
    ctypedef AMF_RESULT (*BUFFER_GETPROPERTY)(AMFBuffer* pThis, const wchar_t *pName, AMFVariantStruct* pValue)
    ctypedef amf_bool (*BUFFER_HASPROPERTY)(AMFBuffer* pThis, const wchar_t *pName, AMFVariantStruct* pValue)
    ctypedef amf_size (*BUFFER_GETPROPERTYCOUNT)(AMFBuffer* pThis)
    ctypedef AMF_MEMORY_TYPE (*BUFFER_GETMEMORYTYPE)(AMFBuffer* pThis)
    ctypedef AMF_DATA_TYPE (*BUFFER_GETDATATYPE)(AMFBuffer* pThis)
    ctypedef amf_bool (*BUFFER_ISREUSABLE)(AMFBuffer* pThis)
    ctypedef void (*BUFFER_SETPTS)(AMFBuffer* pThis, amf_pts pts)
    ctypedef amf_pts (*BUFFER_GETPTS)(AMFBuffer* pThis)
    ctypedef void (*BUFFER_SETDURATION)(AMFBuffer* pThis, amf_pts duration)
    ctypedef amf_pts (*BUFFER_GETDURATION)(AMFBuffer* pThis)
    # AMFBuffer interface
    ctypedef AMF_RESULT (*BUFFER_SETSIZE)(AMFBuffer* pThis, amf_size newSize)
    ctypedef AMF_RESULT (*BUFFER_GETSIZE)(AMFBuffer* pThis)
    ctypedef void* (*BUFFER_GETNATIVE)(AMFBuffer* pThis)

    ctypedef struct AMFBufferVtbl:
        BUFFER_ACQUIRE Acquire
        BUFFER_RELEASE Release
        BUFFER_QUERYINTERFACE QueryInterface
        BUFFER_SETPROPERTY SetProperty
        BUFFER_GETPROPERTY GetProperty
        BUFFER_GETPROPERTYCOUNT GetPropertyCount
        BUFFER_GETMEMORYTYPE GetMemoryType
        BUFFER_GETDATATYPE GetDataType
        BUFFER_ISREUSABLE IsReusable
        BUFFER_SETPTS SetPts
        BUFFER_GETPTS GetPts
        BUFFER_SETDURATION SetDuration
        BUFFER_GETDURATION GetDuration
        BUFFER_SETSIZE SetSize
        BUFFER_GETSIZE GetSize
        BUFFER_GETNATIVE GetNative

    ctypedef struct AMFBuffer:
        const AMFBufferVtbl *pVtbl


cdef inline PLANE_TYPE_STR(AMF_PLANE_TYPE ptype):
    return {
        AMF_PLANE_UNKNOWN: "UNKNOWN",
        AMF_PLANE_PACKED: "PACKED",
        AMF_PLANE_Y: "Y",
        AMF_PLANE_UV: "UV",
        AMF_PLANE_U: "U",
        AMF_PLANE_V: "V",
    }.get(ptype, "unknown")


cdef extern from "core/Trace.h":
    cdef int AMF_TRACE_ERROR
    cdef int AMF_TRACE_WARNING
    cdef int AMF_TRACE_INFO
    cdef int AMF_TRACE_DEBUG
    cdef int AMF_TRACE_TRACE

    ctypedef void (*TRACEWRITE)(AMFTraceWriter* pThis, const wchar_t* scope, const wchar_t* message) noexcept nogil
    ctypedef void (*TRACEFLUSH)(AMFTraceWriter* pThis) noexcept nogil

    ctypedef struct AMFTraceWriterVtbl:
        TRACEWRITE Write
        TRACEFLUSH Flush

    ctypedef struct AMFTraceWriter:
        const AMFTraceWriterVtbl *pVtbl

    ctypedef void (*TRACEW)(AMFTrace* pThis, const wchar_t* src_path, amf_int32 line, amf_int32 level, const wchar_t* scope,amf_int32 countArgs, const wchar_t* format, ...)
    ctypedef void (*TRACE)(AMFTrace* pThis, const wchar_t* src_path, amf_int32 line, amf_int32 level, const wchar_t* scope, const wchar_t* message, va_list* pArglist)
    ctypedef amf_int32 (*SETGLOBALLEVEL)(AMFTrace* pThis, amf_int32 level)
    ctypedef amf_int32 (*GETGLOBALLEVEL)(AMFTrace* pThis)
    ctypedef const wchar_t* (*GETRESULTTEXT)(AMFTrace* pThis, AMF_RESULT res)
    ctypedef void (*REGISTERWRITER)(AMFTrace* pThis, const wchar_t* writerID, AMFTraceWriter* pWriter, amf_bool enable)
    ctypedef void (*UNREGISTERWRITER)(AMFTrace* pThis, const wchar_t* writerID)

    ctypedef struct AMFTraceVtbl:
        TRACEW TraceW
        TRACE Trace
        SETGLOBALLEVEL SetGlobalLevel
        GETGLOBALLEVEL GetGlobalLevel
        GETRESULTTEXT GetResultText
        REGISTERWRITER RegisterWriter
        UNREGISTERWRITER UnregisterWriter

    ctypedef struct AMFTrace:
        const AMFTraceVtbl *pVtbl


cdef extern from "core/Surface.h":
    ctypedef enum AMF_FRAME_TYPE:
        AMF_FRAME_STEREO_FLAG
        AMF_FRAME_LEFT_FLAG
        AMF_FRAME_RIGHT_FLAG
        AMF_FRAME_BOTH_FLAG
        AMF_FRAME_INTERLEAVED_FLAG
        AMF_FRAME_FIELD_FLAG
        AMF_FRAME_EVEN_FLAG
        AMF_FRAME_ODD_FLAG

        # values
        AMF_FRAME_UNKNOWN
        AMF_FRAME_PROGRESSIVE

        AMF_FRAME_INTERLEAVED_EVEN_FIRST
        AMF_FRAME_INTERLEAVED_ODD_FIRST
        AMF_FRAME_FIELD_SINGLE_EVEN
        AMF_FRAME_FIELD_SINGLE_ODD

        AMF_FRAME_STEREO_LEFT
        AMF_FRAME_STEREO_RIGHT
        AMF_FRAME_STEREO_BOTH

        AMF_FRAME_INTERLEAVED_EVEN_FIRST_STEREO_LEFT
        AMF_FRAME_INTERLEAVED_EVEN_FIRST_STEREO_RIGHT
        AMF_FRAME_INTERLEAVED_EVEN_FIRST_STEREO_BOTH

        AMF_FRAME_INTERLEAVED_ODD_FIRST_STEREO_LEFT
        AMF_FRAME_INTERLEAVED_ODD_FIRST_STEREO_RIGHT
        AMF_FRAME_INTERLEAVED_ODD_FIRST_STEREO_BOTH

    ctypedef enum AMF_SURFACE_FORMAT:
        AMF_SURFACE_NV12    # 1  - planar 4:2:0 Y width x height + packed UV width/2 x height/2 - 8 bit per component
        AMF_SURFACE_YV12    # 2  - planar 4:2:0 Y width x height + V width/2 x height/2 + U width/2 x height/2 - 8 bit per component
        AMF_SURFACE_BGRA    # 3  - packed 4:4:4 - 8 bit per component
        AMF_SURFACE_ARGB    # 4  - packed 4:4:4 - 8 bit per component
        AMF_SURFACE_RGBA    # 5  - packed 4:4:4 - 8 bit per component
        AMF_SURFACE_GRAY8   # 6  - single component - 8 bit
        AMF_SURFACE_YUV420P # 7  - planar 4:2:0 Y width x height + U width/2 x height/2 + V width/2 x height/2 - 8 bit per component
        AMF_SURFACE_U8V8    # 8  - packed double component - 8 bit per component
        AMF_SURFACE_YUY2    # 9  - packed 4:2:2 Byte 0=8-bit Y'0; Byte 1=8-bit Cb; Byte 2=8-bit Y'1; Byte 3=8-bit Cr
        AMF_SURFACE_P010    # 10 - planar 4:2:0 Y width x height + packed UV width/2 x height/2 - 10 bit per component (16 allocated, upper 10 bits are used)
        AMF_SURFACE_RGBA_F16 # 11 - packed 4:4:4 - 16 bit per component float
        AMF_SURFACE_UYVY    # 12 - packed 4:2:2 the similar to YUY2 but Y and UV swapped: Byte 0=8-bit Cb; Byte 1=8-bit Y'0; Byte 2=8-bit Cr Byte 3=8-bit Y'1; (used the same DX/CL/Vulkan storage as YUY2)
        AMF_SURFACE_R10G10B10A2  # 13 - packed 4:4:4 to 4 bytes, 10 bit per RGB component, 2 bits per A
        AMF_SURFACE_Y210    # 14 - packed 4:2:2 - Word 0=10-bit Y'0; Word 1=10-bit Cb; Word 2=10-bit Y'1; Word 3=10-bit Cr
        AMF_SURFACE_AYUV    # 15 - packed 4:4:4 - 8 bit per component YUVA
        AMF_SURFACE_Y410    # 16 - packed 4:4:4 - 10 bit per YUV component, 2 bits per A, AVYU
        AMF_SURFACE_Y416    # 17 - packed 4:4:4 - 16 bit per component 4 bytes, AVYU
        AMF_SURFACE_GRAY32  # 18 - single component - 32 bit
        AMF_SURFACE_P012    # 19 - planar 4:2:0 Y width x height + packed UV width/2 x height/2 - 12 bit per component (16 allocated, upper 12 bits are used)
        AMF_SURFACE_P016

    ctypedef AMF_RESULT (*SURFACE_SETPROPERTY)(AMFSurface* pThis, const wchar_t* name, AMFVariantStruct value)
    ctypedef amf_long (*SURFACE_ACQUIRE)(AMFSurface* pThis)
    ctypedef amf_long (*SURFACE_RELEASE)(AMFSurface* pThis)
    ctypedef AMF_SURFACE_FORMAT (*SURFACE_GETFORMAT)(AMFSurface* pThis)
    ctypedef amf_size (*SURFACE_GETPLANESCOUNT)(AMFSurface* pThis)
    ctypedef AMFPlane* (*SURFACE_GETPLANEAT)(AMFSurface* pThis, amf_size index)
    ctypedef AMFPlane* (*SURFACE_GETPLANE)(AMFSurface* pThis, AMF_PLANE_TYPE type)
    ctypedef AMF_FRAME_TYPE (*SURFACE_GETFRAMETYPE)(AMFSurface* pThis)
    ctypedef AMF_RESULT (*SURFACE_CONVERT)(AMFSurface* pThis, AMF_MEMORY_TYPE type)
    ctypedef AMF_RESULT (*SURFACE_INTEROP)(AMFSurface* pThis, AMF_MEMORY_TYPE type)
    ctypedef AMF_RESULT (*SETCROP)(amf_int32 x,amf_int32 y, amf_int32 width, amf_int32 height)
    ctypedef AMF_RESULT (*COPYSURFACEREGION)(AMFSurface* pDest, amf_int32 dstX, amf_int32 dstY, amf_int32 srcX, amf_int32 srcY, amf_int32 width, amf_int32 height)

    ctypedef struct AMFSurfaceVtbl:
        SURFACE_SETPROPERTY SetProperty
        SURFACE_ACQUIRE Acquire
        SURFACE_RELEASE Release
        SURFACE_GETFORMAT GetFormat
        SURFACE_GETPLANESCOUNT GetPlanesCount
        SURFACE_GETPLANEAT GetPlaneAt
        SURFACE_GETPLANE GetPlane
        SURFACE_GETFRAMETYPE GetFrameType
        SURFACE_CONVERT Convert
        SURFACE_INTEROP Interop
        SETCROP SetCrop
        COPYSURFACEREGION CopySurfaceRegion

    ctypedef struct AMFSurface:
        const AMFSurfaceVtbl *pVtbl


cdef inline SURFACE_FORMAT_STR(AMF_SURFACE_FORMAT fmt):
    return {
        AMF_SURFACE_NV12: "NV12",
        AMF_SURFACE_YV12: "YV12",
        AMF_SURFACE_BGRA: "BGRA",
        AMF_SURFACE_ARGB: "ARGB",
        AMF_SURFACE_RGBA: "RGBA",
        AMF_SURFACE_GRAY8: "GRAY8",
        AMF_SURFACE_YUV420P: "YUV420P",
        AMF_SURFACE_U8V8: "U8V8",
        AMF_SURFACE_YUY2: "YUY2",
        AMF_SURFACE_P010: "P010",
        AMF_SURFACE_RGBA_F16: "RGBA_F16",
        AMF_SURFACE_UYVY: "UYVY",
        AMF_SURFACE_R10G10B10A2: "R10G10B10A2",
        AMF_SURFACE_Y210: "Y210",
        AMF_SURFACE_AYUV: "AYUV",
        AMF_SURFACE_Y410: "Y410",
        AMF_SURFACE_Y416: "Y416",
        AMF_SURFACE_GRAY32: "GRAY32",
        AMF_SURFACE_P012: "P012",
        AMF_SURFACE_P016: "P016",
    }.get(fmt, "unknown")


cdef inline DATA_TYPE_STR(AMF_DATA_TYPE dtype):
    return {
        AMF_DATA_BUFFER: "BUFFER",
        AMF_DATA_SURFACE: "SURFACE",
        AMF_DATA_AUDIO_BUFFER: "AUDIO_BUFFER",
        AMF_DATA_USER: "USER",
    }.get(dtype, "unknown")


cdef inline FRAME_TYPE_STR(AMF_FRAME_TYPE ftype):
    return {
        AMF_FRAME_STEREO_FLAG: "STEREO_FLAG",
        AMF_FRAME_LEFT_FLAG: "LEFT_FLAG",
        AMF_FRAME_RIGHT_FLAG: "RIGHT_FLAG",
        AMF_FRAME_BOTH_FLAG: "BOTH_FLAG",
        AMF_FRAME_INTERLEAVED_FLAG: "INTERLEAVED_FLAG",
        AMF_FRAME_FIELD_FLAG: "FIELD_FLAG",
        AMF_FRAME_EVEN_FLAG: "EVEN_FLAG",
        AMF_FRAME_ODD_FLAG: "ODD_FLAG",

        # values
        AMF_FRAME_UNKNOWN: "UNKNOWN",
        AMF_FRAME_PROGRESSIVE: "PROGRESSIVE",

        AMF_FRAME_INTERLEAVED_EVEN_FIRST: "INTERLEAVED_EVEN_FIRST",
        AMF_FRAME_INTERLEAVED_ODD_FIRST: "INTERLEAVED_ODD_FIRST",
        AMF_FRAME_FIELD_SINGLE_EVEN: "FIELD_SINGLE_EVEN",
        AMF_FRAME_FIELD_SINGLE_ODD: "FIELD_SINGLE_ODD",

        AMF_FRAME_STEREO_LEFT: "STEREO_LEFT",
        AMF_FRAME_STEREO_RIGHT: "STEREO_RIGHT",
        AMF_FRAME_STEREO_BOTH: "STEREO_BOTH",

        AMF_FRAME_INTERLEAVED_EVEN_FIRST_STEREO_LEFT: "INTERLEAVED_EVEN_FIRST_STEREO_LEFT",
        AMF_FRAME_INTERLEAVED_EVEN_FIRST_STEREO_RIGHT: "INTERLEAVED_EVEN_FIRST_STEREO_RIGHT",
        AMF_FRAME_INTERLEAVED_EVEN_FIRST_STEREO_BOTH: "INTERLEAVED_EVEN_FIRST_STEREO_BOTH",

        AMF_FRAME_INTERLEAVED_ODD_FIRST_STEREO_LEFT: "INTERLEAVED_ODD_FIRST_STEREO_LEFT",
        AMF_FRAME_INTERLEAVED_ODD_FIRST_STEREO_RIGHT: "INTERLEAVED_ODD_FIRST_STEREO_RIGHT",
        AMF_FRAME_INTERLEAVED_ODD_FIRST_STEREO_BOTH: "INTERLEAVED_ODD_FIRST_STEREO_BOTH",
    }.get(ftype, "unknown")


cdef inline ACCEL_TYPE_STR(AMF_ACCELERATION_TYPE dtype):
    return{
        AMF_ACCEL_NOT_SUPPORTED: "not supported",
        AMF_ACCEL_HARDWARE: "hardware",
        AMF_ACCEL_GPU: "gpu",
        AMF_ACCEL_SOFTWARE: "software",
    }.get(dtype, "unknown")


cdef extern from "core/Context.h":
    ctypedef enum AMF_DX_VERSION:
        AMF_DX9     # 90
        AMF_DX9_EX  # 91
        AMF_DX11_0  # 110
        AMF_DX11_1  # 111
        AMF_DX12    # 120

    ctypedef amf_long (*CONTEXT_ACQUIRE)(AMFContext* pThis)
    ctypedef amf_long (*CONTEXT_RELEASE)(AMFContext* pThis)
    ctypedef AMF_RESULT (*CONTEXT_QUERYINTERFACE)(AMFContext* pThis, const AMFGuid *interfaceID, void** ppInterface)

    ctypedef AMF_RESULT (*CONTEXT_TERMINATE)(AMFContext *context)
    ctypedef AMF_RESULT (*CONTEXT_INITDX11)(AMFContext *context, void* pDX11Device, AMF_DX_VERSION dxVersionRequired)
    ctypedef void* (*CONTEXT_GETDX11DEVICE)(AMFContext *context, AMF_DX_VERSION dxVersionRequired)
    ctypedef AMF_RESULT (*CONTEXT_LOCKDX11)(AMFContext *context)
    ctypedef AMF_RESULT (*CONTEXT_UNLOCKDX11)(AMFContext *context)

    ctypedef AMF_RESULT (*CONTEXT_INITOPENGL)(AMFContext *context, amf_handle hOpenGLContext, amf_handle hWindow, amf_handle hDC)
    ctypedef amf_handle (*CONTEXT_GETOPENGLCONTEXT)(AMFContext *context)
    ctypedef amf_handle (*CONTEXT_GETOPENGLDRAWABLE)(AMFContext *context)
    ctypedef AMF_RESULT (*CONTEXT_LOCKOPENGL)(AMFContext *context)
    ctypedef AMF_RESULT (*CONTEXT_UNLOCKOPENGL)(AMFContext *context)

    ctypedef AMF_RESULT (*CONTEXT_ALLOCSURFACE)(AMFContext *context, AMF_MEMORY_TYPE type, AMF_SURFACE_FORMAT format, amf_int32 width, amf_int32 height, AMFSurface** ppSurface)

    # AMF_RESULT          AMF_STD_CALL AllocBuffer(AMF_MEMORY_TYPE type, amf_size size, AMFBuffer** ppBuffer) = 0
    # AMF_RESULT          AMF_STD_CALL CreateBufferFromHostNative(void* pHostBuffer, amf_size size, AMFBuffer** ppBuffer, AMFBufferObserver* pObserver) = 0
    # AMF_RESULT          AMF_STD_CALL CreateSurfaceFromHostNative(AMF_SURFACE_FORMAT format, amf_int32 width, amf_int32 height, amf_int32 hPitch, amf_int32 vPitch, void* pData,
    #                                                 AMFSurface** ppSurface, AMFSurfaceObserver* pObserver) = 0
    # AMF_RESULT          AMF_STD_CALL CreateSurfaceFromDX11Native(void* pDX11Surface, AMFSurface** ppSurface, AMFSurfaceObserver* pObserver) = 0
    # AMF_RESULT          AMF_STD_CALL CreateSurfaceFromOpenGLNative(AMF_SURFACE_FORMAT format, amf_handle hGLTextureID, AMFSurface** ppSurface, AMFSurfaceObserver* pObserver) = 0
    # AMF_RESULT          AMF_STD_CALL CreateSurfaceFromGrallocNative(amf_handle hGrallocSurface, AMFSurface** ppSurface, AMFSurfaceObserver* pObserver) = 0

    ctypedef struct AMFContextVtbl:
        CONTEXT_ACQUIRE Acquire
        CONTEXT_RELEASE Release
        CONTEXT_QUERYINTERFACE QueryInterface

        CONTEXT_TERMINATE Terminate
        CONTEXT_INITDX11 InitDX11
        CONTEXT_GETDX11DEVICE GetDX11Device
        CONTEXT_LOCKDX11 LockDX11
        CONTEXT_UNLOCKDX11 UnlockDX11

        CONTEXT_INITOPENGL InitOpenGL
        CONTEXT_GETOPENGLCONTEXT GetOpenGLContext
        CONTEXT_GETOPENGLDRAWABLE GetOpenGLDrawable
        CONTEXT_LOCKOPENGL LockOpenGL
        CONTEXT_UNLOCKOPENGL UnlockOpenGL

        CONTEXT_ALLOCSURFACE AllocSurface

    ctypedef struct AMFContext:
        const AMFContextVtbl *pVtbl

    ctypedef AMF_RESULT (*CONTEXT1_INITVULKAN)(AMFContext1 *context, void* pVulkanDevice)
    ctypedef void* (*CONTEXT1_GETVULKANDEVICE)(AMFContext1 *context)
    ctypedef AMF_RESULT (*CONTEXT1_LOCKVULKAN)(AMFContext1 *context)
    ctypedef AMF_RESULT (*CONTEXT1_UNLOCKVULKAN)(AMFContext1 *context)

    ctypedef struct AMFContext1Vtbl:
        CONTEXT1_INITVULKAN InitVulkan
        CONTEXT1_GETVULKANDEVICE GetVulkanDevice
        CONTEXT1_LOCKVULKAN LockVulkan
        CONTEXT1_UNLOCKVULKAN UnlockVulkan

    ctypedef struct AMFContext1:
        const AMFContext1Vtbl *pVtbl


cdef extern from "components/ComponentCaps.h":
    ctypedef enum AMF_ACCELERATION_TYPE:
        AMF_ACCEL_NOT_SUPPORTED
        AMF_ACCEL_HARDWARE
        AMF_ACCEL_GPU
        AMF_ACCEL_SOFTWARE

    ctypedef void (*GETWIDTHRANGE)(AMFIOCaps* pThis, amf_int32* minWidth, amf_int32* maxWidth)
    ctypedef void (*GETHEIGHTRANGE)(AMFIOCaps* pThis, amf_int32* minHeight, amf_int32* maxHeight)
    # Get memory alignment in lines: Vertical aligmnent should be multiples of this number
    ctypedef amf_int32 (*GETVERTALIGN)(AMFIOCaps* pThis)
    # Enumerate supported surface pixel formats
    ctypedef amf_int32 (*GETNUMOFFORMATS)(AMFIOCaps* pThis)
    ctypedef AMF_RESULT (*GETFORMATAT)(AMFIOCaps* pThis, amf_int32 index, AMF_SURFACE_FORMAT* format, amf_bool* native)
    # Enumerate supported memory types
    ctypedef amf_int32 (*GETNUMOFMEMORYTYPES)(AMFIOCaps* pThis)
    ctypedef AMF_RESULT (*GETMEMORYTYPEAT)(AMFIOCaps* pThis, amf_int32 index, AMF_MEMORY_TYPE* memType, amf_bool* native)
    ctypedef amf_bool (*ISINTERLACEDSUPPORTED)(AMFIOCaps* pThis)

    ctypedef struct AMFIOCapsVtbl:
        GETWIDTHRANGE GetWidthRange
        GETHEIGHTRANGE GetHeightRange
        GETVERTALIGN GetVertAlign
        GETNUMOFFORMATS GetNumOfFormats
        GETFORMATAT GetFormatAt
        GETNUMOFMEMORYTYPES GetNumOfMemoryTypes
        GETMEMORYTYPEAT GetMemoryTypeAt
        ISINTERLACEDSUPPORTED IsInterlacedSupported

    ctypedef struct AMFIOCaps:
        const AMFIOCapsVtbl *pVtbl

    ctypedef AMF_RESULT (*CAPSSETPROPERTY)(AMFCaps* pThis, const wchar_t* name, AMFVariantStruct value)
    ctypedef AMF_RESULT (*CAPSGETPROPERTY)(AMFCaps* pThis, const wchar_t* name, AMFVariantStruct* pValue)
    ctypedef AMF_RESULT (*HASPROPERTY)(AMFCaps* pThis, const wchar_t* name)

    ctypedef AMF_ACCELERATION_TYPE (*GETACCELERATIONTYPE)(AMFCaps* pThis)
    ctypedef AMF_RESULT (*GETINPUTCAPS)(AMFCaps* pThis, AMFIOCaps** input)
    ctypedef AMF_RESULT (*GETOUTPUTCAPS)(AMFCaps* pThis, AMFIOCaps** output)

    ctypedef struct AMFCapsVtbl:
        CAPSSETPROPERTY SetProperty
        CAPSGETPROPERTY GetProperty
        HASPROPERTY HasProperty
        GETACCELERATIONTYPE GetAccelerationType
        GETINPUTCAPS GetInputCaps
        GETOUTPUTCAPS GetOutputCaps

    ctypedef struct AMFCaps:
        const AMFCapsVtbl *pVtbl


cdef extern from "components/Component.h":
    ctypedef amf_long (*COMPONENT_ACQUIRE)(AMFComponent* pThis)
    ctypedef amf_long (*COMPONENT_RELEASE)(AMFComponent* pThis)

    ctypedef AMF_RESULT (*COMPONENT_SETPROPERTY)(AMFComponent* pThis, const wchar_t *pName, AMFVariantStruct value)
    ctypedef AMF_RESULT (*COMPONENT_GETPROPERTY)(AMFComponent* pThis, const wchar_t *pName, AMFVariantStruct* pValue)
    ctypedef amf_bool (*COMPONENT_HASPROPERTY)(AMFComponent* pThis, const wchar_t *pName, AMFVariantStruct* pValue)
    ctypedef amf_size (*COMPONENT_GETPROPERTYCOUNT)(AMFComponent* pThis)

    ctypedef AMF_RESULT (*COMPONENT_INIT)(AMFComponent* pThis, AMF_SURFACE_FORMAT format,amf_int32 width,amf_int32 height)
    ctypedef AMF_RESULT (*COMPONENT_REINIT)(AMFComponent* pThis, amf_int32 width,amf_int32 height)
    ctypedef AMF_RESULT (*COMPONENT_TERMINATE)(AMFComponent* pThis)
    ctypedef AMF_RESULT (*COMPONENT_DRAIN)(AMFComponent* pThis)
    ctypedef AMF_RESULT (*COMPONENT_FLUSH)(AMFComponent* pThis)
    ctypedef AMF_RESULT (*COMPONENT_SUBMITINPUT)(AMFComponent* pThis, AMFData* pData)
    ctypedef AMF_RESULT (*COMPONENT_QUERYOUTPUT)(AMFComponent* pThis, AMFData** ppData)
    ctypedef AMFContext* (*COMPONENT_GETCONTEXT)(AMFComponent* pThis)
    ctypedef AMF_RESULT (*COMPONENT_SETOUTPUTDATAALLOCATORCB)(AMFComponent* pThis, AMFDataAllocatorCB* callback)
    ctypedef AMF_RESULT (*COMPONENT_GETCAPS)(AMFComponent* pThis, AMFCaps** ppCaps)
    ctypedef AMF_RESULT (*COMPONENT_OPTIMIZE)(AMFComponent* pThis, AMFComponentOptimizationCallback* pCallback)

    ctypedef struct AMFComponentVtbl:
        # AMFInterface interface
        COMPONENT_ACQUIRE Acquire
        COMPONENT_RELEASE Release

        # AMFPropertyStorage
        COMPONENT_SETPROPERTY SetProperty
        COMPONENT_GETPROPERTY GetProperty
        COMPONENT_HASPROPERTY HasProperty
        COMPONENT_GETPROPERTYCOUNT GetPropertyCount

        # AMFComponent interface
        COMPONENT_INIT Init
        COMPONENT_REINIT ReInit
        COMPONENT_TERMINATE Terminate
        COMPONENT_DRAIN Drain
        COMPONENT_FLUSH Flush
        COMPONENT_SUBMITINPUT SubmitInput
        COMPONENT_QUERYOUTPUT QueryOutput
        COMPONENT_GETCONTEXT GetContext
        COMPONENT_SETOUTPUTDATAALLOCATORCB SetOutputDataAllocatorCB
        COMPONENT_GETCAPS GetCaps
        COMPONENT_OPTIMIZE Optimize

    ctypedef struct AMFComponent:
        const AMFComponentVtbl *pVtbl


cdef extern from "core/Factory.h":
    ctypedef struct AMFDebug:
        pass
    ctypedef struct AMFPrograms:
        pass

    ctypedef AMF_RESULT (*AMFCREATECONTEXT)(AMFFactory *factory, AMFContext** context)
    ctypedef AMF_RESULT (*AMFCREATECOMPONENT)(AMFFactory *factory,AMFContext* context, const wchar_t* id, AMFComponent** ppComponent)
    ctypedef AMF_RESULT (*AMFSETCACHEFOLDER)(AMFFactory *factory,const wchar_t *path)
    ctypedef const wchar_t (*AMFGETCACHEFOLDER)(AMFFactory *factory)
    ctypedef AMF_RESULT (*AMFGETDEBUG)(AMFFactory *factory, AMFDebug **debug)
    ctypedef AMF_RESULT (*AMFGETTRACE)(AMFFactory *factory, AMFTrace **trace)
    ctypedef AMF_RESULT (*AMFGETPROGRAMS)(AMFFactory *factory, AMFPrograms **programs)

    ctypedef struct AMFFactoryVtbl:
        AMFCREATECONTEXT CreateContext
        AMFCREATECOMPONENT CreateComponent
        AMFSETCACHEFOLDER SetCacheFolder
        AMFGETCACHEFOLDER GetCacheFolder
        AMFGETDEBUG GetDebug
        AMFGETTRACE GetTrace
        AMFGETPROGRAMS GetPrograms

    ctypedef struct AMFFactory:
        const AMFFactoryVtbl *pVtbl


cdef extern from "components/VideoEncoderVCE.h":
    enum AMF_VIDEO_ENCODER_USAGE_ENUM:
        AMF_VIDEO_ENCODER_USAGE_TRANSCODING
        AMF_VIDEO_ENCODER_USAGE_ULTRA_LOW_LATENCY
        AMF_VIDEO_ENCODER_USAGE_LOW_LATENCY
        AMF_VIDEO_ENCODER_USAGE_WEBCAM
        AMF_VIDEO_ENCODER_USAGE_HIGH_QUALITY
        AMF_VIDEO_ENCODER_USAGE_LOW_LATENCY_HIGH_QUALITY

    enum AMF_VIDEO_ENCODER_QUALITY_PRESET_ENUM:
        AMF_VIDEO_ENCODER_QUALITY_PRESET_BALANCED
        AMF_VIDEO_ENCODER_QUALITY_PRESET_SPEED
        AMF_VIDEO_ENCODER_QUALITY_PRESET_QUALITY

    enum AMF_VIDEO_ENCODER_PROFILE_ENUM:
        AMF_VIDEO_ENCODER_PROFILE_BASELINE
        AMF_VIDEO_ENCODER_PROFILE_MAIN
        AMF_VIDEO_ENCODER_PROFILE_HIGH
        AMF_VIDEO_ENCODER_PROFILE_CONSTRAINED_BASELINE
        AMF_VIDEO_ENCODER_PROFILE_CONSTRAINED_HIGH

    enum AMF_VIDEO_ENCODER_H264_LEVEL_ENUM:
        AMF_H264_LEVEL__1
        AMF_H264_LEVEL__1_1
        AMF_H264_LEVEL__1_2
        AMF_H264_LEVEL__1_3
        AMF_H264_LEVEL__2
        AMF_H264_LEVEL__2_1
        AMF_H264_LEVEL__2_2
        AMF_H264_LEVEL__3
        AMF_H264_LEVEL__3_1
        AMF_H264_LEVEL__3_2
        AMF_H264_LEVEL__4
        AMF_H264_LEVEL__4_1
        AMF_H264_LEVEL__4_2
        AMF_H264_LEVEL__5
        AMF_H264_LEVEL__5_1
        AMF_H264_LEVEL__5_2
        AMF_H264_LEVEL__6
        AMF_H264_LEVEL__6_1
        AMF_H264_LEVEL__6_2

    enum AMF_VIDEO_ENCODER_PICTURE_STRUCTURE_ENUM:
        AMF_VIDEO_ENCODER_PICTURE_STRUCTURE_NONE
        AMF_VIDEO_ENCODER_PICTURE_STRUCTURE_FRAME
        AMF_VIDEO_ENCODER_PICTURE_STRUCTURE_TOP_FIELD
        AMF_VIDEO_ENCODER_PICTURE_STRUCTURE_BOTTOM_FIELD

    enum AMF_VIDEO_ENCODER_OUTPUT_DATA_TYPE_ENUM:
        AMF_VIDEO_ENCODER_OUTPUT_DATA_TYPE_IDR
        AMF_VIDEO_ENCODER_OUTPUT_DATA_TYPE_I
        AMF_VIDEO_ENCODER_OUTPUT_DATA_TYPE_P
        AMF_VIDEO_ENCODER_OUTPUT_DATA_TYPE_B

    enum AMF_VIDEO_ENCODER_OUTPUT_BUFFER_TYPE_ENUM:
        AMF_VIDEO_ENCODER_OUTPUT_BUFFER_TYPE_FRAME
        AMF_VIDEO_ENCODER_OUTPUT_BUFFER_TYPE_TILE
        AMF_VIDEO_ENCODER_OUTPUT_BUFFER_TYPE_TILE_LAST

    enum AMF_VIDEO_ENCODER_FORCE_PICTURE_TYPE:
        AMF_VIDEO_ENCODER_PICTURE_TYPE_NONE
        AMF_VIDEO_ENCODER_PICTURE_TYPE_SKIP
        AMF_VIDEO_ENCODER_PICTURE_TYPE_IDR
        AMF_VIDEO_ENCODER_PICTURE_TYPE_I
        AMF_VIDEO_ENCODER_PICTURE_TYPE_P
        AMF_VIDEO_ENCODER_PICTURE_TYPE_B


cdef inline AMF_VIDEO_ENCODER_OUTPUT_DATA_TYPE_STR(AMF_VIDEO_ENCODER_OUTPUT_DATA_TYPE_ENUM dtype):
    return {
        AMF_VIDEO_ENCODER_OUTPUT_DATA_TYPE_IDR: "IDR",
        AMF_VIDEO_ENCODER_OUTPUT_DATA_TYPE_I: "I",
        AMF_VIDEO_ENCODER_OUTPUT_DATA_TYPE_P: "P",
        AMF_VIDEO_ENCODER_OUTPUT_DATA_TYPE_B: "B",
    }.get(dtype, "unknown")


cdef extern from "components/VideoEncoderHEVC.h":
    enum AMF_VIDEO_ENCODER_HEVC_USAGE_ENUM:
        AMF_VIDEO_ENCODER_HEVC_USAGE_TRANSCODING
        AMF_VIDEO_ENCODER_HEVC_USAGE_ULTRA_LOW_LATENCY
        AMF_VIDEO_ENCODER_HEVC_USAGE_LOW_LATENCY
        AMF_VIDEO_ENCODER_HEVC_USAGE_WEBCAM
        AMF_VIDEO_ENCODER_HEVC_USAGE_HIGH_QUALITY
        AMF_VIDEO_ENCODER_HEVC_USAGE_LOW_LATENCY_HIGH_QUALITY

    enum AMF_VIDEO_ENCODER_HEVC_QUALITY_PRESET_ENUM:
        AMF_VIDEO_ENCODER_HEVC_QUALITY_PRESET_QUALITY
        AMF_VIDEO_ENCODER_HEVC_QUALITY_PRESET_BALANCED
        AMF_VIDEO_ENCODER_HEVC_QUALITY_PRESET_SPEED

    enum AMF_VIDEO_ENCODER_HEVC_PROFILE_ENUM:
        AMF_VIDEO_ENCODER_HEVC_PROFILE_MAIN
        AMF_VIDEO_ENCODER_HEVC_PROFILE_MAIN_10

    enum AMF_VIDEO_ENCODER_HEVC_TIER_ENUM:
        AMF_VIDEO_ENCODER_HEVC_TIER_MAIN
        AMF_VIDEO_ENCODER_HEVC_TIER_HIGH

    enum AMF_VIDEO_ENCODER_LEVEL_ENUM:
        AMF_LEVEL_1
        AMF_LEVEL_2
        AMF_LEVEL_2_1
        AMF_LEVEL_3
        AMF_LEVEL_3_1
        AMF_LEVEL_4
        AMF_LEVEL_4_1
        AMF_LEVEL_5
        AMF_LEVEL_5_1
        AMF_LEVEL_5_2
        AMF_LEVEL_6
        AMF_LEVEL_6_1
        AMF_LEVEL_6_2


cdef extern from "components/VideoEncoderAV1.h":
    enum AMF_VIDEO_ENCODER_AV1_ENCODING_LATENCY_MODE_ENUM:
        AMF_VIDEO_ENCODER_AV1_ENCODING_LATENCY_MODE_NONE
        AMF_VIDEO_ENCODER_AV1_ENCODING_LATENCY_MODE_POWER_SAVING_REAL_TIME
        AMF_VIDEO_ENCODER_AV1_ENCODING_LATENCY_MODE_REAL_TIME
        AMF_VIDEO_ENCODER_AV1_ENCODING_LATENCY_MODE_LOWEST_LATENCY

    enum AMF_VIDEO_ENCODER_AV1_USAGE_ENUM:
        AMF_VIDEO_ENCODER_AV1_USAGE_TRANSCODING
        AMF_VIDEO_ENCODER_AV1_USAGE_ULTRA_LOW_LATENCY
        AMF_VIDEO_ENCODER_AV1_USAGE_LOW_LATENCY
        AMF_VIDEO_ENCODER_AV1_USAGE_WEBCAM
        AMF_VIDEO_ENCODER_AV1_USAGE_HIGH_QUALITY
        AMF_VIDEO_ENCODER_AV1_USAGE_LOW_LATENCY_HIGH_QUALITY

    enum AMF_VIDEO_ENCODER_AV1_PROFILE_ENUM:
        AMF_VIDEO_ENCODER_AV1_PROFILE_MAIN

    enum AMF_VIDEO_ENCODER_AV1_LEVEL_ENUM:
        AMF_VIDEO_ENCODER_AV1_LEVEL_2_0
        AMF_VIDEO_ENCODER_AV1_LEVEL_2_1
        AMF_VIDEO_ENCODER_AV1_LEVEL_2_2
        AMF_VIDEO_ENCODER_AV1_LEVEL_2_3
        AMF_VIDEO_ENCODER_AV1_LEVEL_3_0
        AMF_VIDEO_ENCODER_AV1_LEVEL_3_1
        AMF_VIDEO_ENCODER_AV1_LEVEL_3_2
        AMF_VIDEO_ENCODER_AV1_LEVEL_3_3
        AMF_VIDEO_ENCODER_AV1_LEVEL_4_0
        AMF_VIDEO_ENCODER_AV1_LEVEL_4_1
        AMF_VIDEO_ENCODER_AV1_LEVEL_4_2
        AMF_VIDEO_ENCODER_AV1_LEVEL_4_3
        AMF_VIDEO_ENCODER_AV1_LEVEL_5_0
        AMF_VIDEO_ENCODER_AV1_LEVEL_5_1
        AMF_VIDEO_ENCODER_AV1_LEVEL_5_2
        AMF_VIDEO_ENCODER_AV1_LEVEL_5_3
        AMF_VIDEO_ENCODER_AV1_LEVEL_6_0
        AMF_VIDEO_ENCODER_AV1_LEVEL_6_1
        AMF_VIDEO_ENCODER_AV1_LEVEL_6_2
        AMF_VIDEO_ENCODER_AV1_LEVEL_6_3
        AMF_VIDEO_ENCODER_AV1_LEVEL_7_0
        AMF_VIDEO_ENCODER_AV1_LEVEL_7_1
        AMF_VIDEO_ENCODER_AV1_LEVEL_7_2
        AMF_VIDEO_ENCODER_AV1_LEVEL_7_3

    enum AMF_VIDEO_ENCODER_AV1_QUALITY_PRESET_ENUM:
        AMF_VIDEO_ENCODER_AV1_QUALITY_PRESET_HIGH_QUALITY
        AMF_VIDEO_ENCODER_AV1_QUALITY_PRESET_QUALITY
        AMF_VIDEO_ENCODER_AV1_QUALITY_PRESET_BALANCED
        AMF_VIDEO_ENCODER_AV1_QUALITY_PRESET_SPEED

    enum AMF_VIDEO_ENCODER_AV1_ALIGNMENT_MODE_ENUM:
        AMF_VIDEO_ENCODER_AV1_ALIGNMENT_MODE_64X16_ONLY
        AMF_VIDEO_ENCODER_AV1_ALIGNMENT_MODE_64X16_1080P_CODED_1082
        AMF_VIDEO_ENCODER_AV1_ALIGNMENT_MODE_NO_RESTRICTIONS
